/*
** File: portability.c
** Project: ADMXRC3 API module library
** Purpose: VxWorks-specific portability functions used by ADMXRC3 API module library
**
** (C) Copyright Alpha Data 2010
*/

#include "platform.h"
#include "../portability.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <limits.h>

portable_file_t
portableFileOpenA(
  const char* pFileName,
  bool_t bWrite)
{
  return fopen(pFileName, bWrite ? "wb" : "rb");
}

void
portableFileClose(
  portable_file_t hFile)
{
  fclose(hFile);
}

bool_t
portableFileGetSize(
  portable_file_t hFile,
  uint64_t* pSize)
{
  int fd;
  struct stat info;

  fd = fileno(hFile);
  if (fstat(fd, &info)) {
    return FALSE;
  }
  *pSize = info.st_size;
  return TRUE;
}

bool_t
portableFileRead(
  portable_file_t hFile,
  void* pBuffer,
  uint64_t n,
  uint64_t* pNumRead)
{
  size_t chunk, numRead;
  uint64_t totalRead;

  totalRead = 0;
  while (n) {
    if (n > (size_t)-1) {
      chunk = (size_t)-1;
    } else {
      chunk = (size_t)n;
    }
    numRead = fread(pBuffer, 1, chunk, hFile);
    if (numRead == 0) {
      break;
    }
    totalRead += numRead;
    n -= numRead;
  }
  *pNumRead = totalRead;
  return TRUE;
}

bool_t
portableFileSetPointer(
  portable_file_t hFile,
  uint64_t offset)
{
  if (offset > LONG_MAX) {
    return FALSE;
  }
  if (fseek(hFile, (long)offset, SEEK_SET)) {
    return FALSE;
  }
  return TRUE;
}

size_t
portableStringLength(
  char* pDst,
  size_t max)
{
  char c;
  size_t n = 0;

  if (!max) {
    return 0;
  }
  max--;
  while (max) {
    c = *pDst++;
    if ('\0' == c) {
      break;
    }
    max--;
    n++;
  }
  return n;
}

void
portableStringCat(
  char* pDst,
  size_t max,
  const char* pSrc)
{
  size_t n;

  if (!max) {
    return;
  }
  n = portableStringLength(pDst, max);
  max--;
  max -= n;
  pDst += n;
  if (max) {
    memcpy(pDst, pSrc, max);
  }
  pDst[max] = '\0';
}

void
portableStringCopy(
  char* pDst,
  size_t max,
  const char* pSrc)
{
  char c;

  if (!max) {
    return;
  }
  max--;
  while (max) {
    c = *pSrc;
    if ('\0' == c) {
      break;
    }
    *pDst++ = *pSrc++;
    max--;
  }
  *pDst++ = '\0';
}

void
portableStringLower(
  char* pString,
  size_t bufSize)
{
  char c;

  if (!bufSize) {
    return;
  }
  bufSize--;
  while (bufSize) {
    c = *pString;
    if ('\0' == c) {
      break;
    }
    *pString++ = tolower(c);
    bufSize--;
  }
}

void
portableStringUpper(
  char* pString,
  size_t bufSize)
{
  char c;

  if (!bufSize) {
    return;
  }
  bufSize--;
  while (bufSize) {
    c = *pString;
    if ('\0' == c) {
      break;
    }
    *pString++ = toupper(c);
    bufSize--;
  }
}

int
portableStringCompareA(
  const char* pString1,
  const char* pString2)
{
  return strcmp(pString1, pString2);
}
