/*
** File: ics843034_01.c  
** Project: ADM-XRC core driver
** Purpose: Functions for programming the ICS843034-01 clock generator.
**
** (C) Copyright Alpha Data 2008-2009
*/

#include <df.h>

#include "device.h"
#include "ics843034_01.h"

/*
** Bits 0-16 of the returned Clock Word are the serial clock programming word
** for the ICS843034_01, but bit 17 is the reference clock select, which is
** driven to the SEL pins of the ICS843034-01 by ADB1-XRC4FX.
*/
ClockWordStatus
clockWordIcs843034_01(
	Adb3CoreDeviceContext* pDevCtx,
  uint32_t flags,
	uint32_t xtalFrequency1,
	uint32_t xtalFrequency2,
	uint64_t frequencyReq,
	uint64_t frequencyMin,
	uint64_t frequencyMax,
  uint32_t* pClockWord,
	uint64_t* pActualFrequency)
{
	uint32_t pDiv = 1; /* P divide is 1 because P pin is floating */
	uint32_t T = 0x0; /* Make TEST output low */
	const uint32_t nVal[8] = { 1, 2, 3, 4, 5, 6, 8, 16 };
	uint32_t xtal, bestXtal = 0; /* init. to avoid compiler whine */
	uint32_t mMin, mMax;
	uint32_t m, bestM = 0, n, nIdx, bestNIdx  = 0; /* init. to avoid compiler whine */
	uint32_t minFrequency, maxFrequency;
	uint32_t refFrequency;
	uint32_t f, bestFrequency = 0; /* init. to avoid compiler whine */
	uint32_t error, bestError = (uint32_t)-1;

	dfDebugPrint(3, ("clockWordXrc4fxLclk: xtalFrequency1=%lu xtalFrequency2=%lu reqFrequency=%llu\n",
    (unsigned long)xtalFrequency1, (unsigned long)xtalFrequency2, (unsigned long long)frequencyReq));

  if (xtalFrequency2 == 0 || xtalFrequency1 == 0) {
		return ClockWordGeneralFailure;
  }

	for (xtal = 0; xtal < 2; xtal++) {
		refFrequency = xtal ? xtalFrequency2 : xtalFrequency1;
		mMin = (490000000 * pDiv + refFrequency - 1) / refFrequency;
		mMax = 640000000 * pDiv / refFrequency;
		minFrequency = refFrequency * mMin / (16 * pDiv) - 1000;
		maxFrequency = refFrequency * mMax / pDiv + 1000;
		if (frequencyReq < minFrequency || frequencyReq > maxFrequency) {
			continue;
		}
		for (m = mMin; m <= mMax; m++) {
			for (nIdx = 0; nIdx < 8; nIdx++) {
				n = nVal[nIdx];
				f = (uint32_t)refFrequency * m / (n * pDiv);
        if ((flags & CoreClockMinimumFrequency) && f < frequencyMin) {
          continue;
        }
        if ((flags & CoreClockMaximumFrequency) && f > frequencyMax) {
          continue;
        }
				if (f > (uint32_t)frequencyReq) {
					error = f - (uint32_t)frequencyReq;
				} else {
					error = (uint32_t)frequencyReq - f;
				}
				if (error < bestError) {
					bestError = error;
					bestFrequency = f;
					bestM = m;
					bestNIdx = nIdx;
					bestXtal = xtal;
				}
			}
		}
	}

	if (bestM == 0) {
		/* Failed to find a suitable value */
		return ClockWordOutOfRange;
	}

	*pActualFrequency = bestFrequency;
	*pClockWord = (bestXtal << 17) | (T << 15) | (bestNIdx << 12) | (bestNIdx << 9) | (bestM << 0);

	dfDebugPrint(3, ("clockWordIcs843034_01: bestFrequency=%lu clockWord=0x%08lx\n", (unsigned long)bestFrequency, (unsigned long)*pClockWord));

	return ClockWordSuccess;
}
