/*
** File: main.c  
** Project: ADB3 core driver
** Purpose: OS-independent entry point into ADB3 core driver
**
** (C) Copyright Alpha Data 2009-2010
*/

#include <df.h>

#include "device.h"
#include "coreclock.h"
#include "coredma.h"
#include "coreflash.h"
#include "corefpga.h"
#include "coreif.h"
#include "coremisc.h"
#include "coremodule.h"
#include "corenotify.h"
#include "coresensor.h"
#include "corevpd.h"
#include "flash.h"
#include "flash_cfi.h"
#include "hwmon.h"
#include "identify.h"
#include "interrupt.h"
#include "resource.h"

#if !ADB3_MONOLITHIC
const char g_dfDriverName[] = "adb3core";
#endif

static Adb3CoreInterface g_coreInterface;

static void
coreTestDebugMsg(
  void* pDevice,
  int level,
  const char* pMessage)
{
  dfDebugPrint(level, ("%s", pMessage));
}

static void
dumpRawResources(
  DfBusType busType,
  DfBusResources* pResources)
{
  unsigned int i, n;
  DfPciIrqLine* pIrq;
  DfPciBar* pBar;

  switch (busType) {
  case DfBusPci:
    n = pResources->variant.pci.numMemBar;
    for (i = 0; i < n; i++) {
      pBar = &pResources->variant.pci.memBar[i];
      dfDebugPrint(0, ("Memory BAR%u base=0x%08lx_%08lx size=0x%08lx_%08lx flags=0x%x\n",
        i, dfSplitUint64(pBar->base), dfSplitUint64(pBar->size), (unsigned int)pBar->flags));
    }
    n = pResources->variant.pci.numIoBar;
    for (i = 0; i < n; i++) {
      pBar = &pResources->variant.pci.ioBar[i];
      dfDebugPrint(0, ("IO     BAR%u base=0x%08lx_%08lx size=0x%08lx_%08lx flags=0x%x\n",
        i, dfSplitUint64(pBar->base), dfSplitUint64(pBar->size), (unsigned int)pBar->flags));
    }
    n = pResources->variant.pci.numIrq;
    for (i = 0; i < n; i++) {
      pIrq = &pResources->variant.pci.irq[i];
      dfDebugPrint(0, ("Interrupt %u bMsgSignalled=%s bLevel=%s line=0x%x pin=0x%x\n",
        i, pIrq->bMessageSignalled ? "TRUE" : "FALSE", pIrq->bLevelSensitive ? "TRUE" : "FALSE", (unsigned int)pIrq->variant.raw.intLine, (unsigned int)pIrq->variant.raw.intPin));
    }
    break;

  default:
    dfDebugPrint(0, ("(unknown bus type)\n"));
    break;
  }
}

static void
freeDeviceContext(
  Adb3CoreDeviceContext* pDevCtx)
{
  unsigned int i;

  if (pDevCtx->bCoreInterfaceRegd) {
    dfUnexportInterface(pDevCtx->pDevObj, g_coreInterfaceName);
    pDevCtx->bCoreInterfaceRegd = FALSE;
    dfDebugPrint(1, ("freeDeviceContext: unregistered core interface\n"));
  }

  dfTimerUninit(&pDevCtx->hwMon.polling.timer);
  for (i = 0; i < MAX_NUM_CLOCK_GENERATOR; i++) {
    ClockGeneratorContext* pDevClockCtx = &pDevCtx->clockGenerator[i];

    dfTimerUninit(&pDevClockCtx->program.timer);
  }
}

static void
initResourceQueue(
  ResourceQueue* pQ)
{
  pQ->pHead = pQ->pTail = NULL;
  pQ->length = 0;
  pQ->users = 0;
}

static void*
enableInterruptsAtomic(
  DfInterruptObject* pInterruptObject, /* ignored */
  void* pArgument1,
  void* pArgument2)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pArgument1;
  boolean_t bEnable = (boolean_t)(uintptr_t)pArgument2;

  pDevCtx->methods.pEnableInterrupts(pDevCtx, bEnable);
  pDevCtx->interrupt.bEnabled = bEnable;

  return NULL; /* ignored */
}

static void
stopDevice(
  DfDeviceObject* pDevObj,
  Adb3CoreDeviceContext* pDevCtx)
{
  unsigned int i, j;
  DfSpinLockFlags f;

  /* Free I/O module VPD buffers */
  for (i = 0; i < MAX_NUM_IO_MODULE; i++) {
    if (NULL != pDevCtx->info.ioModule[i].pVpdBuffer) {
      dfFree(pDevCtx->info.ioModule[i].pVpdBuffer);
      pDevCtx->info.ioModule[i].pVpdBuffer = NULL;
    }
  }

  /* Stop hardware monitor polling timer, if running */
  dfEventInit(&pDevCtx->hwMon.polling.shutdownEvent);
  f = dfSpinLockGet(&pDevCtx->hwMon.lock);
  pDevCtx->hwMon.polling.bShutdown = TRUE;
  if (pDevCtx->hwMon.polling.bTimerRunning) {
    dfSpinLockPut(&pDevCtx->hwMon.lock, f);
    if (!dfTimerCancel(&pDevCtx->hwMon.polling.timer)) {
      dfEventWait(&pDevCtx->hwMon.polling.shutdownEvent);
    }
  } else {
    dfSpinLockPut(&pDevCtx->hwMon.lock, f);
  }
  dfEventUninit(&pDevCtx->hwMon.polling.shutdownEvent);

  /* Disable interrupts in hardware */
  if (pDevCtx->interrupt.bEnabled) {
    dfInterruptExecuteAtomic(&pDevCtx->interrupt.object, enableInterruptsAtomic, (void*)pDevCtx, (void*)(uintptr_t)FALSE);
    pDevCtx->interrupt.bEnabled = FALSE;
    dfDebugPrint(1, ("stopDevice: disabled interrupts\n"));
  }

  /* Free common buffers for DMA descriptors */
  for (i = 0; i < MAX_NUM_DMA_ENGINE; i++) {
    for (j = 0; j < 2; j++) {
      if (NULL != pDevCtx->dma.channel[i].descriptor[j].pDmaBuffer) {
        dfDmaBufferFree(pDevCtx->dma.channel[i].descriptor[j].pDmaBuffer);
        pDevCtx->dma.channel[i].descriptor[j].pDmaBuffer = NULL;
      }
    }
  }

  /* Free DMA mapper */
  if (pDevCtx->bDmaMapperAllocated) {
    dfDmaMapperFree(pDevObj);
    pDevCtx->bDmaMapperAllocated = FALSE;
    dfDebugPrint(1, ("stopDevice: freed DMA mapper\n"));
  }

  /* Detach interrupt handler */
  if (pDevCtx->interrupt.bConnected) {
    dfInterruptDisconnect(pDevCtx->pDevObj, &pDevCtx->interrupt.object);
    pDevCtx->interrupt.bConnected = FALSE;
    dfDebugPrint(1, ("stopDevice: disconnected interrupt handler\n"));
  }

  /* Free Flash information structures */
  if (NULL != pDevCtx->info.flash.pDetail) {
    flashCfiCleanup(pDevCtx->info.flash.pDetail);
    pDevCtx->info.flash.pDetail = NULL;
  }

  /* Free Flash merge buffer */
  if (NULL != pDevCtx->flashControl.pMergeBuffer) {
    dfFree(pDevCtx->flashControl.pMergeBuffer);
    pDevCtx->flashControl.pMergeBuffer = NULL;
  }

  /* Deinitialize hardware */
  if (pDevCtx->hardware.bInitialized) {
    if (!pDevCtx->methods.pInitHardware(pDevCtx, FALSE)) {
      dfDebugPrint(0, ("*** stopDevice: failed to uninitialize hardware\n"));
    } else {
      dfDebugPrint(1, ("stopDevice: uninitialized hardware\n"));
    }
    pDevCtx->hardware.bInitialized = FALSE;
  }

  /* Unmap device from kernel address space */
  if (pDevCtx->hardware.bMapped) {
    if (!pDevCtx->methods.pMapBusResources(pDevCtx, NULL, NULL, TRUE)) {
      dfDebugPrint(0, ("*** stopDevice: failed to unmap bus resources\n"));
    } else {
      pDevCtx->hardware.bMapped = FALSE;
      dfDebugPrint(1, ("stopDevice: unmapped bus resources\n"));
    }
  }
}

boolean_t
adb3CoreOnStart(
  DfDeviceObject* pDevObj,
  Adb3CoreDeviceContext* pDevCtx,
  DfBusResources* pRawRes,
  DfBusResources* pTranslatedRes)
{
  CfiInformation* pDetail;
  uint32_t dmaMapLength;
  unsigned int i, j;

  dfDebugPrint(1, ("adb3CoreOnStart: entered, pDevObj=0x%p pDevCtx=0x%p\n", pDevObj, pDevCtx));

  dfCopyKK(&pDevCtx->rawResources, pRawRes, sizeof(DfBusResources));
  dfCopyKK(&pDevCtx->translatedResources, pTranslatedRes, sizeof(DfBusResources));

  if (dfDebugLevel(2)) {
    dfDebugPrint(0, ("adb3CoreOnStart: raw resources:\n"));
    dumpRawResources(pDevObj->hardwareId.busType, &pDevCtx->rawResources);
  }

  /* Check that device's base address registers are OK */
  if (!pDevCtx->methods.pValidateBusResources(pDevCtx, &pDevCtx->rawResources, &pDevCtx->translatedResources)) {
    dfDebugPrint(0, ("*** adb3CoreOnStart: failed to validate bus resources\n"));
    goto failed;
  }
  dfDebugPrint(1, ("adb3CoreOnStart: validated bus resources\n"));

  /* Map device into kernel address space */
  if (!pDevCtx->methods.pMapBusResources(pDevCtx, &pDevCtx->rawResources, &pDevCtx->translatedResources, FALSE)) {
    dfDebugPrint(0, ("*** adb3CoreOnStart: failed to map bus resources\n"));
    goto failed;
	}
  pDevCtx->hardware.bMapped = TRUE;
  dfDebugPrint(1, ("adb3CoreOnStart: mapped bus resources\n"));

  /* Initialize hardware */
  pDevCtx->info.flash.bPresent = FALSE;
  if (!pDevCtx->methods.pInitHardware(pDevCtx, TRUE)) {
    dfDebugPrint(0, ("*** adb3CoreOnStart: failed to initialize hardware\n"));
    goto failed;
  }
  pDevCtx->hardware.bInitialized = TRUE;
  dfDebugPrint(1, ("adb3CoreOnStart: initialized hardware\n"));

  pDetail = pDevCtx->info.flash.pDetail;
  if (pDevCtx->info.flash.bPresent && NULL != pDetail) {
    if (pDetail->validation.largestBlock > (size_t)-1) {
      dfDebugPrint(0, ("*** adb3CoreOnStart: required Flash merge buffer too large, size=0x%08lx_%08lx\n",
        dfSplitUint64(pDetail->validation.largestBlock)));
      goto failed;
    }
    pDevCtx->flashControl.pMergeBuffer = (uint8_t*)dfMalloc(
      (size_t)pDetail->validation.largestBlock); /* Cast safe due to previous check */
    if (NULL == pDevCtx->flashControl.pMergeBuffer) {
      dfDebugPrint(0, ("*** adb3CoreOnStart: failed to allocate Flash merge buffer\n"));
      goto failed;
    }
    dfDebugPrint(1, ("adb3CoreOnStart: allocated Flash merge buffer, size=0x%08lx_%08lx\n",
      dfSplitUint64(pDetail->validation.largestBlock)));
  }

  if (pDevCtx->info.vpd.bValid && pDevCtx->info.vpd.length) {
    size_t n;

    dfDebugPrint(2, ("adb3CoreOnStart: dump of raw VPD information (max 0x100 bytes):\n"));
    n = pDevCtx->info.vpd.length;
    if (n > 0x100U) {
      n = 0x100U;
    }
    dfDebugDumpMemory(2, 4, 16, 1, FALSE, &pDevCtx->info.vpd.data.generic, n);
  }

  for (i = 0; i < pDevCtx->info.bootstrap.numIoModule; i++) {
    CoreIoModuleInfo* pInfo = &pDevCtx->info.ioModule[i];
    size_t n;

    if (pInfo->bPresent && NULL != pInfo->pVpdBuffer) {
      n = pInfo->vpdLength;
      if (n > 0x100U) {
        n = 0x100U;
      }
      dfDebugPrint(2, ("adb3CoreOnStart: dump of FRU ROM %u information (max 0x100 bytes):\n", i));
      dfDebugDumpMemory(2, 4, 16, 1, FALSE, pInfo->pVpdBuffer, n);
    }
  }

  /* Attach interrupt handler */
  if (NULL != pDevCtx->methods.pIsr) {
    switch (pDevCtx->pDevObj->hardwareId.busType) {
    case DfBusPci:
      if (pDevCtx->translatedResources.variant.pci.numIrq == 1) {
        if (!dfInterruptConnect(pDevCtx->pDevObj, &pDevCtx->interrupt.object, &pDevCtx->translatedResources, DfBusPci, 0, pDevCtx->methods.pIsr, pDevCtx)) {
          dfDebugPrint(0, ("*** adb3CoreOnStart: failed to connect interrupt handler\n"));
          goto failed;
        }
        pDevCtx->interrupt.bConnected = TRUE;
        dfDebugPrint(1, ("adb3CoreOnStart: connected interrupt handler\n"));
      } else {
        dfDebugPrint(1, ("adb3CoreOnStart: not connecting interrupt handler due to wrong number of interrupts (%lu)\n",
          (unsigned long)pDevCtx->translatedResources.variant.pci.numIrq));
      }
      break;

    default:
      dfDebugPrint(1, ("adb3CoreOnStart: not connecting interrupt handler due to unknown bus type (%lu)\n",
        (unsigned long)pDevCtx->pDevObj->hardwareId.busType));
      break;
    }
  } else {
    dfDebugPrint(1, ("adb3CoreOnStart: not connecting interrupt handler due to no ISR method\n"));
  }

  if (pDevCtx->info.bootstrap.numDmaChannel > 0) {
    /* Allocate DMA mapper */
    if (DfParameterSuccess != dfParameterGetUint32(pDevObj->pDrvObj, "DmaMapLength", &dmaMapLength)) {
      dmaMapLength = DEFAULT_DMA_MAP_LENGTH;
    }
    if (dmaMapLength < 1U) {
      dmaMapLength = 1U;
    }
    if (dmaMapLength > (g_dfDmaBufferMaxLength / pDevCtx->info.bootstrap.dmaDescriptorSize)) {
      dmaMapLength = (g_dfDmaBufferMaxLength / pDevCtx->info.bootstrap.dmaDescriptorSize);
    }
    if (!dfDmaMapperAllocate(
      pDevCtx->pDevObj,
      pDevCtx->info.bootstrap.numDmaChannel * 2,
      (unsigned int)dmaMapLength,
      pDevCtx->info.bootstrap.b64BitDmaAddress,
      pDevCtx->info.bootstrap.dmaTransferMaxSize,
      &pDevCtx->dma.mapLength))
    {
      dfDebugPrint(0, ("*** adb3CoreOnStart: failed to allocate DMA mapper\n"));
      goto failed;
    }
    pDevCtx->bDmaMapperAllocated = TRUE;
    dfDebugPrint(1, ("adb3CoreOnStart: allocated DMA mapper, dmaMapLength=%lu\n", (unsigned long)pDevCtx->dma.mapLength));

    /* Allocate common buffers for DMA descriptors, 2 per DMA engine */
    for (i = 0; i < pDevCtx->info.bootstrap.numDmaChannel; i++) {
      for (j = 0; j < 2; j++) {
        /* Request 'dmaDescriptorSize - 1' extra bytes to enable alignment of the returned buffer */
        unsigned int size = pDevCtx->info.bootstrap.dmaDescriptorSize * pDevCtx->dma.mapLength;
        DfDmaBuffer* pDmaBuffer;

        pDmaBuffer = dfDmaBufferAllocate(pDevObj, size, pDevCtx->info.bootstrap.dmaDescriptorSize);
        if (NULL == pDmaBuffer) {
          dfDebugPrint(0, ("*** adb3CoreOnStart: failed to allocate DMA descriptor buffer %lu.%lu\n", (unsigned long)i, (unsigned long)j));
          goto failed;
        }
        pDevCtx->dma.channel[i].descriptor[j].pDmaBuffer = pDmaBuffer;
        dfDebugPrint(1, ("adb3CoreOnStart: DMA descriptor buffer %lu.%lu pKernAddr=0x%p busAddr=0x%08lx_%08lx size=%lu(0x%lx)\n",
          (unsigned long)i, (unsigned long)j, (void*)pDmaBuffer->pKernelAddress,
          dfSplitUint64(pDmaBuffer->busAddress), (unsigned long)size, (unsigned long)size));
        pDevCtx->dma.channel[i].descriptor[j].buffer.pGeneric = pDmaBuffer->pKernelAddress;
        pDevCtx->dma.channel[i].descriptor[j].length = size;
        pDevCtx->dma.channel[i].descriptor[j].busAddress = pDmaBuffer->busAddress;
      }
    }
  }

  /* Enable interrupt in hardware */
  if (pDevCtx->interrupt.bConnected) {
    dfInterruptExecuteAtomic(&pDevCtx->interrupt.object, enableInterruptsAtomic, (void*)pDevCtx, (void*)(uintptr_t)TRUE);
    dfDebugPrint(1, ("adb3CoreOnStart: enabled interrupts\n"));
  }

  return TRUE;

failed:
  stopDevice(pDevObj, pDevCtx);
  return FALSE;
}

void
adb3CoreOnStop(
  DfDeviceObject* pDevObj,
  Adb3CoreDeviceContext* pDevCtx)
{
  dfDebugPrint(1, ("adb3CoreOnStop: entered, pDevObj=0x%p pDevCtx=0x%p\n", pDevObj, pDevCtx));

  stopDevice(pDevObj, pDevCtx);
}

boolean_t
adb3CoreOnAdd(
  DfDeviceObject* pDevObj,
  Adb3CoreDeviceContext* pDevCtx)
{
  unsigned int i, j;

  dfDebugPrint(1, ("adb3CoreOnAdd: entered, pDevObj=0x%p pDevCtx=0x%p\n", pDevObj, pDevCtx));
  
  /* Initialize the device context */
  DF_ZERO_OBJECT(pDevCtx);
  pDevCtx->pDevObj = pDevObj;
  pDevCtx->bCoreInterfaceRegd = FALSE;
  pDevCtx->bDmaMapperAllocated = FALSE;

  /* Initialize hardware stuff */
  pDevCtx->hardware.bridge.pGeneric = NULL;
  pDevCtx->hardware.model.pGeneric = NULL;
  pDevCtx->hardware.pBridged = NULL;
  pDevCtx->hardware.pBridgedPrefetch = NULL;
  pDevCtx->hardware.bInitialized = FALSE;
  pDevCtx->hardware.bMapped = FALSE;

  /* Initialize the methods in the device context */
  pDevCtx->methods.pModelBootstrap = NULL;

  /* Initialize resource queues */
  dfSpinLockInit(&pDevCtx->resource.lock);
  initResourceQueue(&pDevCtx->resource.clockGen.queue);
  for (i = 0; i < MAX_NUM_CLOCK_GENERATOR; i++) {
    pDevCtx->resource.clockGen.users[i] = 0;
  }
  initResourceQueue(&pDevCtx->resource.dmaEngine.queue);
  for (i = 0; i < MAX_NUM_DMA_ENGINE; i++) {
    pDevCtx->resource.dmaEngine.users[i] = 0;
  }
  initResourceQueue(&pDevCtx->resource.eeprom.queue);
  pDevCtx->resource.eeprom.users[0] = 0;
  initResourceQueue(&pDevCtx->resource.flash.queue);
  pDevCtx->resource.flash.users[0] = 0;
  initResourceQueue(&pDevCtx->resource.targetFpga.queue);
  for (i = 0; i < MAX_NUM_TARGET_FPGA; i++) {
    pDevCtx->resource.targetFpga.users[i] = 0;
  }
  initResourceQueue(&pDevCtx->resource.sysMon.queue);
  pDevCtx->resource.sysMon.users[0] = 0;

  /* Initialize Flash stuff */
  pDevCtx->info.flash.pDetail = NULL;
  pDevCtx->info.flash.bPresent = FALSE;
  pDevCtx->info.flash.targetArea.start = 0;
  pDevCtx->info.flash.targetArea.length = 0;
  pDevCtx->flashControl.pMergeBuffer = NULL;

  /* Initialize VPD stuff */
  pDevCtx->info.vpd.bValid = FALSE;
  pDevCtx->info.vpd.bChecksumError = FALSE;

  /* Initialize clock stuff */
  for (i = 0; i < MAX_NUM_CLOCK_GENERATOR; i++) {
    ClockGeneratorContext* pDevClockCtx = &pDevCtx->clockGenerator[i];

    pDevClockCtx->index = i;
    pDevClockCtx->currentFrequency = 0;
    dfTimerInit(&pDevClockCtx->program.timer, pDevCtx);
    pDevClockCtx->program.state = 0;
  }

  /* Initialize interrupt stuff */
  pDevCtx->interrupt.bConnected = FALSE;
  pDevCtx->interrupt.bEnabled = FALSE;
  for (i = 0; i < MAX_NUM_TARGET_FPGA; i++) {
    dfDpcInit(&pDevCtx->interrupt.fpga[i].dpc, &pDevCtx->interrupt.object, fpgaInterruptDpc, pDevCtx);
    dfDpcInit(&pDevCtx->interrupt.hwMon[i].dpc, &pDevCtx->interrupt.object, hwMonAlertDpc, pDevCtx);
  }

  /* Initialize notification stuff */
  for (i = 0; i < MAX_NUM_TARGET_FPGA; i++) {
    dfSpinLockInit(&pDevCtx->notification.fpgaInterrupt[i].lock);
    dfListInit(&pDevCtx->notification.fpgaInterrupt[i].header);
    dfSpinLockInit(&pDevCtx->notification.fpgaAlert[i].lock);
    dfListInit(&pDevCtx->notification.fpgaAlert[i].header);
  }
  dfSpinLockInit(&pDevCtx->notification.sensorPoll.lock);
  dfListInit(&pDevCtx->notification.sensorPoll.header);

  /* Initialize DMA stuff */
  for (i = 0; i < MAX_NUM_DMA_ENGINE; i++) {
    for (j = 0; j < 2; j++) {
      pDevCtx->dma.channel[i].descriptor[j].pDmaBuffer = NULL;
    }
    dfSpinLockInit(&pDevCtx->dma.channel[i].lock);
    dfDpcInit(&pDevCtx->dma.channel[i].completionDpc, &pDevCtx->interrupt.object, dmaCompletedRoutine, (void*)pDevCtx);
    pDevCtx->dma.channel[i].curr.pTicket = NULL;
    pDevCtx->dma.channel[i].curr.state.bCancel = FALSE;
    pDevCtx->dma.channel[i].curr.state.main = 0;
    pDevCtx->dma.channel[i].curr.state.setup.state = 0;
    pDevCtx->dma.channel[i].curr.state.setup.count = 0;
    pDevCtx->dma.channel[i].curr.state.transfer.state = 0;
    pDevCtx->dma.channel[i].curr.state.teardown.state = 0;
    pDevCtx->dma.channel[i].curr.state.teardown.count = 0;
  }

  /* Initialize hardware monitoring stuff */
  dfSpinLockInit(&pDevCtx->hwMon.lock);
  pDevCtx->hwMon.polling.bShutdown = FALSE;
  pDevCtx->hwMon.polling.bTimerRunning = FALSE;
  pDevCtx->hwMon.polling.enable = 0;
  dfTimerInit(&pDevCtx->hwMon.polling.timer, pDevCtx);
  for (i = 0; i < MAX_NUM_SENSOR; i++) {
    pDevCtx->hwMon.history.enable[i] = 0;
  }

  /* Initialize I/O module stuff */
  for (i = 0; i < MAX_NUM_IO_MODULE; i++) {
    pDevCtx->info.ioModule[i].bPresent = FALSE;
    pDevCtx->info.ioModule[i].pVpdBuffer = NULL;
  }

  /* From here on, things may fail */
	dfDebugPrint(1, ("adb3CoreOnAdd: infallible initialization completed\n"));

  if (!identifyDevice(pDevCtx, &pDevObj->hardwareId)) {
    dfDebugPrint(0, ("*** adb3CoreOnAdd: unrecognized hardware ID\n"));
    goto failed;
  }

  if (NULL != pDevCtx->methods.pModelBootstrap) {
    pDevCtx->methods.pModelBootstrap(pDevCtx);
  } else {
    dfDebugPrint(0, ("*** adb3CoreOnAdd: identifyDevice failed to set a ModelBootstrap method\n"));
    goto failed;
  }

  /* Check that bootstrap routine set the appropriate methods */
  if (pDevCtx->info.bootstrap.flashWidth != 0) {
    if (NULL == pDevCtx->methods.pFlash) {
      dfDebugPrint(0, ("*** adb3CoreOnAdd: ModelBootstrap failed to initialize pFlash method\n"));
      goto failed;
    }
  }
  if (pDevCtx->info.bootstrap.numClockGen != 0) {
    if (NULL == pDevCtx->methods.pClockProgram) {
      dfDebugPrint(0, ("*** adb3CoreOnAdd: ModelBootstrap failed to initialize pClockProgram method\n"));
      goto failed;
    }
    if (NULL == pDevCtx->methods.pClockWord) {
      dfDebugPrint(0, ("*** onAdd: ModelBootstrap failed to initialize pClockWord method\n"));
      goto failed;
    }
  }
  if (pDevCtx->info.bootstrap.numTargetFpga != 0) {
    if (NULL == pDevCtx->methods.pFpgaControl) {
      dfDebugPrint(0, ("*** adb3CoreOnAdd: ModelBootstrap failed to initialize pFpgaControl method\n"));
      goto failed;
    }
    if (NULL == pDevCtx->methods.pFpgaSelectMap) {
      dfDebugPrint(0, ("*** adb3CoreOnAdd: ModelBootstrap failed to initialize pFpgaSelectMap method\n"));
      goto failed;
    }
  }
  if (pDevCtx->info.bootstrap.vpdLength != 0) {
    if (NULL == pDevCtx->methods.pReadWriteVpd) {
      dfDebugPrint(0, ("*** adb3CoreOnAdd: ModelBootstrap failed to initialize pReadWriteVpd method\n"));
      goto failed;
    }
  }
  if (pDevCtx->info.bootstrap.numSensor != 0) {
    if (NULL == pDevCtx->methods.pReadSensor) {
      dfDebugPrint(0, ("*** adb3CoreOnAdd: ModelBootstrap failed to initialize pReadSensor method\n"));
      goto failed;
    }
  }
  if (NULL == pDevCtx->methods.pEnableInterrupts ||
    NULL == pDevCtx->methods.pInitHardware ||
    NULL == pDevCtx->methods.pIsr ||
    NULL == pDevCtx->methods.pMapBusResources ||
    NULL == pDevCtx->methods.pValidateBusResources)
  {
    dfDebugPrint(0, ("*** adb3CoreOnAdd: ModelBootstrap failed to initialize all mandatory methods\n"));
    dfDebugPrint(0, ("*** adb3CoreOnAdd:   pEnableInterrupts=0x%p\n", pDevCtx->methods.pEnableInterrupts));
    dfDebugPrint(0, ("*** adb3CoreOnAdd:   pInitHardware=0x%p\n", pDevCtx->methods.pInitHardware));
    dfDebugPrint(0, ("*** adb3CoreOnAdd:   pIsr=0x%p\n", pDevCtx->methods.pIsr));
    dfDebugPrint(0, ("*** adb3CoreOnAdd:   pMapBusResources=0x%p\n", pDevCtx->methods.pMapBusResources));
    dfDebugPrint(0, ("*** adb3CoreOnAdd:   pValidateBusResources=0x%p\n", pDevCtx->methods.pValidateBusResources));
    goto failed;
  }

  dfAssert(pDevCtx->info.bootstrap.numDmaChannel <= MAX_NUM_DMA_ENGINE);
  dfAssert(pDevCtx->info.bootstrap.numClockGen <= MAX_NUM_CLOCK_GENERATOR);
  dfAssert(pDevCtx->info.bootstrap.numTargetFpga <= MAX_NUM_TARGET_FPGA);
  dfAssert(pDevCtx->info.bootstrap.numSensor <= MAX_NUM_SENSOR);
  dfAssert(pDevCtx->info.bootstrap.numWindow <= MAX_NUM_LOCAL_BUS_WINDOW);

  if (!dfExportInterface(pDevObj, g_coreInterfaceName, sizeof(Adb3CoreInterface), &g_coreInterface, pDevCtx)) {
    dfDebugPrint(0, ("*** adb3CoreOnAdd: failed to export core interface\n"));
    goto failed;
  }
  pDevCtx->bCoreInterfaceRegd = TRUE;
  dfDebugPrint(1, ("adb3CoreOnAdd: exported core interface\n"));
  
  return TRUE;
  
 failed:
  freeDeviceContext(pDevCtx);
  return FALSE;
}

void
adb3CoreOnRemove(
  DfDeviceObject* pDevObj,
  Adb3CoreDeviceContext* pDevCtx)
{
  dfDebugPrint(1, ("adb3CoreOnRemove: entered. pDevObj=0x%p pDevCtx=0x%p\n", pDevObj, pDevCtx));
  
  freeDeviceContext(pDevCtx);
}

boolean_t
adb3CoreDriverEntry(
  DfDriverObject* pDrvObj)
{
  dfDebugPrint(1, ("adb3CoreDriverEntry: entered, pDrvObj=0x%p\n", pDrvObj));

  dfAssert(sizeof(ModelRegsAdmxrc6tx) == 0x800U);

  /* Initialize the ADB3 core interface */
  g_coreInterface.pAcquireAsync = coreAcquireAsync;
  g_coreInterface.pAcquireAsyncPoll = coreAcquireAsyncPoll;
  g_coreInterface.pAcquireInitialize = coreAcquireInitialize;
  g_coreInterface.pAcquireSync = coreAcquireSync;
  g_coreInterface.pAcquireCancel = coreAcquireCancel;
  g_coreInterface.pRelease = coreRelease;
  g_coreInterface.pTestDebugMsg = coreTestDebugMsg;
  g_coreInterface.pGetModel = coreGetModel;
  g_coreInterface.pGetNumClockGen = coreGetNumClockGen;
  g_coreInterface.pGetNumDmaChannel = coreGetNumDmaChannel;
  g_coreInterface.pGetNumFlashBank = coreGetNumFlashBank;
  g_coreInterface.pGetNumIoModule = coreGetNumIoModule;
  g_coreInterface.pGetNumSensor = coreGetNumSensor;
  g_coreInterface.pGetNumTargetFpga = coreGetNumTargetFpga;
  g_coreInterface.pGetNumWindow = coreGetNumWindow;
  g_coreInterface.pGetVpdPointer = coreGetVpdPointer;
  g_coreInterface.pGetFlashBlockInfo = coreGetFlashBlockInfo;
  g_coreInterface.pGetFlashInfo = coreGetFlashInfo;
  g_coreInterface.pGetIoModuleInfo = coreGetIoModuleInfo;
  g_coreInterface.pGetSensorInfo = coreGetSensorInfo;
  g_coreInterface.pGetWindowInfo = coreGetWindowInfo;
  g_coreInterface.pGetWindowAddress = coreGetWindowAddress;
  g_coreInterface.pClockTestFreq = coreClockTestFreq;
  g_coreInterface.pNotificationReg = coreNotificationReg;
  g_coreInterface.pNotificationUnreg = coreNotificationUnreg;
  g_coreInterface.pReadSensor = coreReadSensor;
  g_coreInterface.pClockGetFreq = coreClockGetFreq;
  g_coreInterface.pClockProgram = coreClockProgram;
  g_coreInterface.pDmaCancel = coreDmaCancel;
  g_coreInterface.pDmaInitialize = coreDmaInitialize;
  g_coreInterface.pDmaStart = coreDmaStart;
  g_coreInterface.pFlashErase = coreFlashErase;
  g_coreInterface.pFlashRead = coreFlashRead;
  g_coreInterface.pFlashWrite = coreFlashWrite;
  g_coreInterface.pFpgaControl = coreFpgaControl;
  g_coreInterface.pFpgaSelectMap = coreFpgaSelectMap;
  g_coreInterface.pVpdRead = coreVpdRead;
  g_coreInterface.pVpdWrite = coreVpdWrite;

  /* Check that core interface has no NULL pointers */
  {
    const size_t n = sizeof(g_coreInterface) / sizeof(void*);
    void** pp = (void**)&g_coreInterface;
    size_t i;
    boolean_t bError = FALSE;

    for (i = 0; i < n; i++) {
      if (NULL == pp[i]) {
        dfDebugPrint(0, ("*** adb3CoreDriverEntry: core interface entry %ld(0x%lx) is NULL\n", (unsigned long)i, (unsigned long)i));
        bError = TRUE;
      }
    }
    if (bError) {
      return FALSE;
    }
  }

	return TRUE;
}

extern void
adb3CoreDriverExit(
  DfDriverObject* pDrvObj)
{
  dfDebugPrint(1, ("adb3CoreDriverExit: entered, pDrvObj=0x%p\n", pDrvObj));
}

/*
** ---------------------------------------------------------------------------------
** These functions are built into the modular driver, but not the monolithic driver
** ---------------------------------------------------------------------------------
*/

#if !ADB3_MONOLITHIC
static boolean_t
onAdd(
	DfDeviceObject* pDevObj)
{
	Adb3CoreDeviceContext* pDevCtx;

	dfDebugPrint(1, ("onAdd: entered, pDevObj=0x%p\n", pDevObj));

	/* Initialize the device context */
	pDevCtx = (Adb3CoreDeviceContext*)dfDeviceGetContext(pDevObj);
	pDevCtx->pDevObj = pDevObj;

  return adb3CoreOnAdd(pDevObj, pDevCtx);
}
#endif

#if !ADB3_MONOLITHIC
static void
onRemove(
	DfDeviceObject* pDevObj)
{
	Adb3CoreDeviceContext* pDevCtx;

	dfDebugPrint(1, ("onRemove: entered, pDevObj=0x%p\n", pDevObj));

	pDevCtx = (Adb3CoreDeviceContext*)dfDeviceGetContext(pDevObj);
  adb3CoreOnRemove(pDevObj, pDevCtx);
}
#endif

#if !ADB3_MONOLITHIC
static boolean_t
onStart(
	DfDeviceObject* pDevObj,
	DfBusResources* pRawRes,
	DfBusResources* pTranslatedRes)
{
  Adb3CoreDeviceContext* pDevCtx;

  dfDebugPrint(1, ("onStart: entered, pDevObj=0x%p\n", pDevObj));

  pDevCtx = (Adb3CoreDeviceContext*)dfDeviceGetContext(pDevObj);
  return adb3CoreOnStart(pDevObj, pDevCtx, pRawRes, pTranslatedRes);
}
#endif

#if !ADB3_MONOLITHIC
static void
onStop(
	DfDeviceObject* pDevObj)
{
	Adb3CoreDeviceContext* pDevCtx;

	dfDebugPrint(1, ("onStop: entered, pDevObj=0x%p\n", pDevObj));

	pDevCtx = (Adb3CoreDeviceContext*)dfDeviceGetContext(pDevObj);
  adb3CoreOnStop(pDevObj, pDevCtx);
}
#endif

#if !ADB3_MONOLITHIC
boolean_t
dfDriverEntry(
	DfDriverObject* pDrvObj)
{
  dfDebugPrint(1, ("dfDriverEntry: ADB3 Core Driver, pDrvObj=0x%p version=%lu.%lu.%lu.%lu\n",
    pDrvObj,
    (unsigned long)ADB3_VERSION_0, (unsigned long)ADB3_VERSION_1, (unsigned long)ADB3_VERSION_2, (unsigned long)ADB3_VERSION_3));

	pDrvObj->deviceContextSize = sizeof(Adb3CoreDeviceContext);
	pDrvObj->methods.pOnAdd = onAdd;
	pDrvObj->methods.pOnRemove = onRemove;
	pDrvObj->methods.pOnStart = onStart;
	pDrvObj->methods.pOnStop = onStop;

  return adb3coreDriverEntry(pDrvObj);
}
#endif

#if !ADB3_MONOLITHIC
extern void
dfDriverExit(
	DfDriverObject* pDrvObj)
{
	dfDebugPrint(1, ("dfDriverExit: ADB3 Core Driver exiting\n"));

  adb3coreDriverExit(pDrvObj);
}
#endif
