/*
** File: dma.c  
** Project: VxWorks Driver Framework
** Purpose: Implementation of DMA mapper functions
**
** (C) Copyright Alpha Data 2010
*/

#include "df.h"
#include "dfvxworks.h"

#if defined(ADB3_VXBUS_DMA)

DfDmaBuffer*
dfDmaBufferAllocate(
  DfDeviceObject* pDevObj,
  size_t size,
  size_t alignment)
{
  DfDmaBuffer* pDmaBuffer = NULL;
  void* pKernelAddress = NULL;
  uint64_t busAddress = 0;
  VXB_DMA_MAP_ID vxbMapId;
  VXB_DMA_TAG_ID vxbParentTagId, vxbTagId;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);

  vxbParentTagId = vxbDmaBufTagParentGet(pDevObj->system.pVxbDev, 0);
  pDmaBuffer = (DfDmaBuffer*)dfMalloc(sizeof(*pDmaBuffer));
  if (NULL == pDmaBuffer) {
    dfDebugPrint(0, ("*** dfDmaBufferAllocate: failed to allocate DfDmaBuffer object, pDevObj=0x%p\n", pDevObj));
    return NULL;
  }
  pDmaBuffer->size = size;
  pDmaBuffer->system.pDevObj = pDevObj;
  vxbTagId = vxbDmaBufTagCreate(
    pDevObj->system.pVxbDev,        /* VxBus device */
    vxbParentTagId,                 /* parent */
    alignment,                      /* alignment */
    0,                              /* boundary */
    0xFFFFFFFF,                     /* lowaddr */
    0xFFFFFFFF,                     /* highaddr */
    NULL,                           /* filter */
    NULL,                           /* filterarg */
    size,                           /* max size */
    1,                              /* nSegments */
    size,                           /* max seg size */
    VXB_DMABUF_ALLOCNOW | VXB_DMABUF_NOCACHE, /* flags */
    NULL,                           /* lockfunc */
    NULL,                           /* lockarg */
    NULL);                          /* ppDmaTag */
  pDmaBuffer->system.vxbus.tagId = vxbTagId;
  pKernelAddress = vxbDmaBufMemAlloc(pDevObj->system.pVxbDev, vxbTagId, NULL, 0, &vxbMapId);
  if (NULL == pKernelAddress) {
    vxbDmaBufTagDestroy(vxbTagId);
    dfFree(pDmaBuffer);
    dfDebugPrint(0, ("*** dfDmaBufferAllocate: failed to allocate DMA memory, pDevObj=0x%p\n", pDevObj));
    return NULL;
  }
  pDmaBuffer->system.vxbus.mapId = vxbMapId;
  pDmaBuffer->pKernelAddress = pKernelAddress;
  if (OK != vxbDmaBufMapLoad(pDevObj->system.pVxbDev, vxbTagId, vxbMapId, pKernelAddress, size, 0)) {
    vxbDmaBufMapDestroy(vxbTagId, vxbMapId);
    vxbDmaBufTagDestroy(vxbTagId);
    dfFree(pDmaBuffer);
    dfDebugPrint(0, ("*** dfDmaBufferAllocate: failed to load DMA map, pDevObj=0x%p\n", pDevObj));
    return NULL;
  }
  busAddress = (uint64_t)(uintptr_t)pDmaBuffer->system.vxbus.mapId->fragList[0].frag;
  pDmaBuffer->busAddress = busAddress;
  return pDmaBuffer;
}

void
dfDmaBufferFree(
  DfDmaBuffer* pDmaBuffer)
{
  DfDeviceObject* pDevObj;

  pDevObj = pDmaBuffer->system.pDevObj;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);

  vxbDmaBufMapDestroy(pDmaBuffer->system.vxbus.tagId, pDmaBuffer->system.vxbus.mapId);
  vxbDmaBufTagDestroy(pDmaBuffer->system.vxbus.tagId);
  dfFree(pDmaBuffer);
}

boolean_t
dfDmaMapperAllocate(
  DfDeviceObject* pDevObj,
  unsigned int numberOfMaps,
  unsigned int mapLength,          /* In blocks */
  boolean_t b64BitAddressing,
  size_t maxMapSize,
  unsigned int* pActualMapLength)  /* In blocks */
{
  unsigned int i, j;
  DfVxwDmaMapperContext* pMapArray = NULL;
  VXB_DMA_TAG_ID vxbParentTagId;
  VXB_DMA_TAG_ID vxbTagId;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);

  if (pDevObj->system.dma.pMapArray != NULL) {
    dfDebugPrint(0, ("*** dfDmaMapperAllocate: attempt to allocate mapper twice, pDevObj=0x%p\n", pDevObj));
    return FALSE;
  }

  vxbParentTagId = vxbDmaBufTagParentGet(pDevObj->system.pVxbDev, 0);

  pMapArray = (DfVxwDmaMapperContext*)dfMalloc(numberOfMaps * sizeof(DfVxwDmaMapperContext));
  if (NULL == pMapArray) {
    dfDebugPrint(0, ("*** dfDmaMapperAllocate: failed to allocate mapper context array, pDevObj=0x%p\n", pDevObj));
    return FALSE;
  }
  pDevObj->system.dma.pMapArray = pMapArray;
  pDevObj->system.dma.numMap = numberOfMaps;
  pDevObj->system.dma.maxMapRegister = mapLength;
  pDevObj->system.dma.maxMapSize = maxMapSize;
  for (i = 0; i < numberOfMaps; i++) {
    pMapArray[i].pDeviceObject = pDevObj;
    pMapArray[i].index = i;
    pMapArray[i].pTable = NULL;
    pMapArray[i].vxbus.tagId = NULL;
    pMapArray[i].vxbus.pMapIdTable = NULL;
#if DF_DBG_BUILD
    pMapArray[i].curr.bIdle = TRUE;
#endif
  }

  for (i = 0; i < numberOfMaps; i++) {
    pMapArray[i].pTable = (DfDmaMapperNode*)dfMalloc(mapLength * sizeof(DfDmaMapperNode));
    if (NULL == pMapArray[i].pTable) {
      dfDebugPrint(0, ("*** dfDmaMapperAllocate: failed to allocate table for map %lu, pDevObj=0x%p mapLength=%lu",
		       (unsigned long)i, pDevObj, (unsigned long)mapLength));
      goto failed;
    }
    pMapArray[i].vxbus.pMapIdTable = (VXB_DMA_MAP_ID*)dfMalloc(mapLength * sizeof(VXB_DMA_MAP_ID));
    if (NULL == pMapArray[i].vxbus.pMapIdTable) {
      dfDebugPrint(0, ("*** dfDmaMapperAllocate: failed to allocate VXBUS map ID table for map %lu, pDevObj=0x%p mapLength=%lu",
		       (unsigned long)i, pDevObj, (unsigned long)mapLength));
      goto failed;
    }
    for (j = 0; j < mapLength; j++) {
      pMapArray[i].vxbus.pMapIdTable[j] = NULL;
    }
    vxbTagId = vxbDmaBufTagCreate(
      pDevObj->system.pVxbDev,        /* VxBus device */
      vxbParentTagId,                 /* parent */
      1,                              /* alignment */
      0,                              /* boundary */
      0xFFFFFFFF,                     /* lowaddr */
      0xFFFFFFFF,                     /* highaddr */
      NULL,                           /* filter */
      NULL,                           /* filterarg */
      DFVXW_VXBUS_DMA_MAX_SEGMENT_LENGTH, /* max size */
      mapLength,                      /* nSegments */
      DFVXW_VXBUS_DMA_MAX_SEGMENT_LENGTH, /* max seg size */
      VXB_DMABUF_ALLOCNOW,            /* flags */
      NULL,                           /* lockfunc */
      NULL,                           /* lockarg */
      NULL);                          /* ppDmaTag */
    if (NULL == vxbTagId) {
      dfDebugPrint(0, ("*** dfDmaMapperAllocate: failed to create VXBUS DMA tag for map %lu, mapLength=%lu segLength=%lu",
		       (unsigned long)i, pDevObj, (unsigned long)numberOfMaps, (unsigned long)mapLength, (unsigned long)DFVXW_VXBUS_DMA_MAX_SEGMENT_LENGTH));
      goto failed;
    }
    pMapArray[i].vxbus.tagId = vxbTagId;
    for (j = 0; j < mapLength; j++) {
      if (NULL == vxbDmaBufMapCreate(pDevObj->system.pVxbDev, vxbTagId, 0, &pMapArray[i].vxbus.pMapIdTable[j])) {
        dfDebugPrint(0, ("*** dfDmaMapperAllocate: failed to allocate VXBUS map ID %lu for map %lu, pDevObj=0x%p mapLength=%lu",
          (unsigned long)j, (unsigned long)i, pDevObj, (unsigned long)mapLength));
        goto failed;
      }
    }
  }

  *pActualMapLength = mapLength;
  dfDebugPrint(3, ("dfDmaMapperAllocate: pDevObj=0x%p numberOfMaps=%lu mapLength=%lu\n",
		   pDevObj, (unsigned long)numberOfMaps, (unsigned long)mapLength));
  return TRUE;

 failed:
  if (NULL != pMapArray) {
    for (i = 0; i < numberOfMaps; i++) {
      vxbTagId = pMapArray[i].vxbus.tagId;
      if (NULL != vxbTagId && NULL != pMapArray[i].vxbus.pMapIdTable) {
        for (j = 0; j < mapLength; j++) {
          if (NULL != pMapArray[i].vxbus.pMapIdTable[j]) {
            vxbDmaBufMapDestroy(vxbTagId, pMapArray[i].vxbus.pMapIdTable[j]);
            pMapArray[i].vxbus.pMapIdTable[j] = NULL;
          }
        }
      }
      if (NULL != vxbTagId) {
        vxbDmaBufTagDestroy(vxbTagId);
        pMapArray[i].vxbus.tagId = NULL;
      }
      if (NULL != pMapArray[i].vxbus.pMapIdTable) {
        dfFree(pMapArray[i].vxbus.pMapIdTable);
        pMapArray[i].vxbus.pMapIdTable = NULL;
      }
      if (NULL != pMapArray[i].pTable) {
        dfFree(pMapArray[i].vxbus.pMapIdTable);
        pMapArray[i].vxbus.pMapIdTable = NULL;
      }
      dfFree(pMapArray[i].pTable);
      pMapArray[i].pTable = NULL;
    }
    dfFree(pMapArray);
    pDevObj->system.dma.pMapArray = NULL;
  }
  return FALSE;
}

extern void
dfDmaMapperFree(
  DfDeviceObject* pDevObj)
{
  DfVxwDmaMapperContext* pMapArray;
  unsigned int i, j, numberOfMaps, mapLength;
  VXB_DMA_TAG_ID vxbTagId;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);

  pMapArray = pDevObj->system.dma.pMapArray;
  numberOfMaps = pDevObj->system.dma.numMap;
  mapLength = pDevObj->system.dma.maxMapRegister;
  if (NULL != pMapArray) {
    for (i = 0; i < numberOfMaps; i++) {
      vxbTagId = pMapArray[i].vxbus.tagId;
      if (NULL != vxbTagId && NULL != pMapArray[i].vxbus.pMapIdTable) {
        for (j = 0; j < mapLength; j++) {
          if (NULL != pMapArray[i].vxbus.pMapIdTable[j]) {
            vxbDmaBufMapDestroy(vxbTagId, pMapArray[i].vxbus.pMapIdTable[j]);
            pMapArray[i].vxbus.pMapIdTable[j] = NULL;
          }
        }
      }
      if (NULL != vxbTagId) {
        vxbDmaBufTagDestroy(vxbTagId);
        pMapArray[i].vxbus.tagId = NULL;
      }
      if (NULL != pMapArray[i].vxbus.tagId) {
        vxbDmaBufTagDestroy(pMapArray[i].vxbus.tagId);
        pMapArray[i].vxbus.tagId = NULL;
      }
      if (NULL != pMapArray[i].pTable) {
        dfFree(pMapArray[i].pTable);
        pMapArray[i].pTable = NULL;
      }
    }
    dfFree(pMapArray);
    pDevObj->system.dma.pMapArray = NULL;
  } else {
    dfDebugPrint(0, ("*** dfDmaMapperFree: attempting to free mapper when mapper not allocated, pDevObj=0x%p\n", pDevObj));
  }
}

#define PAGE_MASK (0x1000)

extern boolean_t
dfDmaMapperSetup(
  DfDeviceObject* pDevObj,
  unsigned int mapIndex,
  DfBufferDescription* pBufferDescription,
  DfDmaMapperCallback* pCallback,
  void* pCallbackContext,
  boolean_t bWriteToDevice,
  boolean_t bFixedLocalAddress,
  DfDmaMapperSlice* pSlices,
  size_t numSlice,
  DfDmaMapperPosition* pStartPosition)
{
  const unsigned int pageOffsetMask = DFVXW_VXBUS_DMA_MAX_SEGMENT_LENGTH - 1U;
  DfDmaMapperSlice* pSlice;
  DfDmaMapperNode* pNode;
  DfVxwDmaMapperContext* pMapperContext;
  DfDmaMapperPosition newPosition;
  size_t sliceIndex, slicePosition;
  size_t sliceOffset, sliceLength, sliceRemaining = 0;
  uint64_t sliceLocalAddress;
  unsigned int pageOffset, pageRemaining;
  size_t chunk, maxMapSize;
  unsigned int bufferPageOffset;
  unsigned int numEntry, maxEntry;
  uint8_t* pBufferAddress;
  uint8_t* pAddress;
  VXB_DEVICE_ID pVxbDev;
  VXB_DMA_TAG_ID vxbTagId;
  VXB_DMA_MAP_ID* pVxbMapIdTable, vxbMapId;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);
  dfAssert(numSlice > 0);
  dfAssert(mapIndex < pDevObj->system.dma.numMap);

  maxMapSize = pDevObj->system.dma.maxMapSize;

  pVxbDev = pDevObj->system.pVxbDev;

  pBufferAddress = (uint8_t*)pBufferDescription->pAddress;
  bufferPageOffset = (unsigned int)(uintptr_t)pBufferAddress & pageOffsetMask;

  sliceIndex = pStartPosition->sliceIndex;
  dfAssert(sliceIndex < numSlice);
  slicePosition = pStartPosition->slicePosition;

  pMapperContext = &pDevObj->system.dma.pMapArray[mapIndex];
  maxEntry = pDevObj->system.dma.maxMapRegister;
  pNode = pMapperContext->pTable;
  vxbTagId = pMapperContext->vxbus.tagId;
  pVxbMapIdTable = pMapperContext->vxbus.pMapIdTable;

  numEntry = 0;
  while (sliceIndex < numSlice && numEntry < maxEntry) {
    /* Get current offset into current slice, and remaining bytes in current slice */
    pSlice = pSlices + sliceIndex;
    pAddress = pBufferAddress + pSlice->offset + slicePosition;
    sliceOffset = pSlice->offset + slicePosition + bufferPageOffset;
    sliceLength = pSlice->length;
    dfAssert(slicePosition < sliceLength);
    sliceRemaining = sliceLength - slicePosition;
    sliceLocalAddress = pSlice->localAddress;
    if (!bFixedLocalAddress) {
      sliceLocalAddress += slicePosition;
    }
    /* Find corresponding page index in region and region index */
    pageOffset = (unsigned int)sliceOffset & pageOffsetMask;
    pageRemaining = DFVXW_VXBUS_DMA_MAX_SEGMENT_LENGTH - pageOffset;
    /* Calculate how many bytes we can map this time around */
    chunk = maxMapSize;
    if (chunk > pageRemaining) {
      chunk = pageRemaining;
    }
    if (chunk > sliceRemaining) {
      chunk = sliceRemaining;
    }
    /* Map the page chunk */
    vxbMapId = pVxbMapIdTable[numEntry];
    vxbDmaBufMapLoad(pVxbDev, vxbTagId, vxbMapId, pAddress, chunk, 0);
    if (bWriteToDevice) {
      vxbDmaBufSync(pVxbDev, vxbTagId, vxbMapId, VXB_DMABUFSYNC_POSTWRITE);
    }
    dfAssert(vxbMapId->nFrags == 1);
    dfAssert(vxbMapId->fragList[0].fragLen == chunk);
    pNode->busAddress = (uint64_t)(uintptr_t)vxbMapId->fragList[0].frag;
    pNode->length = chunk;
    pNode->localAddress = sliceLocalAddress;
    pNode++;
    /* Increment pointers etc. */
    slicePosition += chunk;
    if (slicePosition == sliceLength) {
      /* Reached end of current slice */
      sliceIndex++;
      slicePosition = 0;
    }
    numEntry++;
  }

  pMapperContext->curr.bWriteToDevice = bWriteToDevice;
  pMapperContext->curr.numEntry = numEntry;
  newPosition.sliceIndex = sliceIndex;
  newPosition.slicePosition = slicePosition;
  
  /* Finished, so call the client's callback function */
  pCallback(
    pDevObj,
    pMapperContext->index,
    pMapperContext->pTable,
    numEntry,
    &newPosition,
    pCallbackContext);

  return TRUE;
}

extern void
dfDmaMapperTeardown(
  DfDeviceObject* pDevObj,
  unsigned int mapIndex)
{
  DfDmaMapperNode* pNode;
  DfVxwDmaMapperContext* pMapperContext;
  unsigned int i, numEntry;
  boolean_t bReadFromDevice;
  VXB_DEVICE_ID pVxbDev;
  VXB_DMA_TAG_ID vxbTagId;
  VXB_DMA_MAP_ID* pVxbMapIdTable, vxbMapId;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);

  pVxbDev = pDevObj->system.pVxbDev;
  pMapperContext = &pDevObj->system.dma.pMapArray[mapIndex];
  pNode = pMapperContext->pTable;
  numEntry = pMapperContext->curr.numEntry;
  bReadFromDevice = !pMapperContext->curr.bWriteToDevice;
  vxbTagId = pMapperContext->vxbus.tagId;
  pVxbMapIdTable = pMapperContext->vxbus.pMapIdTable;

  for (i = 0; i < numEntry; i++) {
    vxbMapId = pVxbMapIdTable[i];
    if (bReadFromDevice) {
      vxbDmaBufSync(pVxbDev, vxbTagId, vxbMapId, VXB_DMABUFSYNC_PREREAD);
    }
    vxbDmaBufMapUnload(vxbTagId, vxbMapId);
    pNode++;
  }
}

#else

DfDmaBuffer*
dfDmaBufferAllocate(
  DfDeviceObject* pDevObj,
  size_t size,
  size_t alignment)
{
  DfDmaBuffer* pDmaBuffer = NULL;
  uint8_t* pKernelAddress;
  uint8_t* pAlignedAddress;
  size_t extraSize, offset;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);
  extraSize = size + alignment - 1;
  dfAssert(extraSize >= size);
  if (size > g_dfDmaBufferMaxLength) {
    dfDebugPrint(0, ("*** dfDmaBufferAllocate: attempt to allocate more than %lu(0x%lx) bytes, pDevObj=0x%p size=%lu(0x%lx) extraSize=%lu(0x%lx)\n",
      (unsigned long)g_dfDmaBufferMaxLength, (unsigned long)g_dfDmaBufferMaxLength, (void*)pDevObj,
      (unsigned long)size, (unsigned long)size, (unsigned long)extraSize, (unsigned long)extraSize));
    return NULL;
  }
  pDmaBuffer = (DfDmaBuffer*)dfMalloc(sizeof(*pDmaBuffer));
  if (NULL == pDmaBuffer) {
    dfDebugPrint(0, ("*** dfDmaBufferAllocate: failed to allocate DfDmaBuffer object, pDevObj=0x%p\n", pDevObj));
    return NULL;
  }
  pKernelAddress = cacheDmaMalloc(extraSize);
  if (NULL == pKernelAddress) {
    dfFree(pDmaBuffer);
    dfDebugPrint(0, ("*** dfDmaBufferAllocate: failed to allocate DMA memory, pDevObj=0x%p\n", (void*)pDevObj));
    return NULL;
  }
  pAlignedAddress = DF_ALIGN_POINTER(pKernelAddress, alignment);
  offset = pAlignedAddress - pKernelAddress;
  pDmaBuffer->size = size;
  pDmaBuffer->pKernelAddress = pKernelAddress + offset;
  pDmaBuffer->system.pDevObj = pDevObj;
  if (!dfVxwLegacyCpuToBusAddress(pDevObj, pDmaBuffer->pKernelAddress, &pDmaBuffer->busAddress)) {
    dfAssert(FALSE);
  }
  pDmaBuffer->system.legacy.unaligned.size = extraSize;
  pDmaBuffer->system.legacy.unaligned.pKernelAddress = pKernelAddress;
  return pDmaBuffer;
}

void
dfDmaBufferFree(
  DfDmaBuffer* pDmaBuffer)
{
  DfDeviceObject* pDevObj;

  pDevObj = pDmaBuffer->system.pDevObj;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);

  cacheDmaFree(pDmaBuffer->system.legacy.unaligned.pKernelAddress);
  dfFree(pDmaBuffer);
}

boolean_t
dfDmaMapperAllocate(
  DfDeviceObject* pDevObj,
  unsigned int numberOfMaps,
  unsigned int mapLength,          /* In blocks */
  boolean_t b64BitAddressing,
  size_t maxMapSize,
  unsigned int* pActualMapLength)  /* In blocks */
{
  unsigned int i;
  DfVxwDmaMapperContext* pMapArray = NULL;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);

  if (pDevObj->system.dma.pMapArray != NULL) {
    dfDebugPrint(0, ("*** dfDmaMapperAllocate: attempt to allocate mapper twice, pDevObj=0x%p\n", pDevObj));
    return FALSE;
  }

  pMapArray = (DfVxwDmaMapperContext*)dfMalloc(numberOfMaps * sizeof(DfVxwDmaMapperContext));
  if (NULL == pMapArray) {
    dfDebugPrint(0, ("*** dfDmaMapperAllocate: failed to allocate mapper context array, pDevObj=0x%p\n", pDevObj));
    return FALSE;
  }
  pDevObj->system.dma.pMapArray = pMapArray;
  pDevObj->system.dma.numMap = numberOfMaps;
  pDevObj->system.dma.maxMapRegister = mapLength;
  pDevObj->system.dma.maxMapSize = maxMapSize;
  for (i = 0; i < numberOfMaps; i++) {
    pMapArray[i].pDeviceObject = pDevObj;
    pMapArray[i].index = i;
    pMapArray[i].pTable = NULL;
#if DF_DBG_BUILD
    pMapArray[i].curr.bIdle = TRUE;
#endif
  }

  for (i = 0; i < numberOfMaps; i++) {
    pMapArray[i].pTable = (DfDmaMapperNode*)dfMalloc(mapLength * sizeof(DfDmaMapperNode));
    if (NULL == pMapArray[i].pTable) {
      dfDebugPrint(0, ("*** dfDmaMapperAllocate: failed to allocate table for map %lu, pDevObj=0x%p mapLength=%lu",
		       (unsigned long)i, pDevObj, (unsigned long)mapLength));
      goto failed;
    }
  }

  *pActualMapLength = mapLength;
  dfDebugPrint(3, ("dfDmaMapperAllocate: pDevObj=0x%p numberOfMaps=%lu mapLength=%lu\n",
		   pDevObj, (unsigned long)numberOfMaps, (unsigned long)mapLength));
  return TRUE;

 failed:
  if (NULL != pMapArray) {
    for (i = 0; i < numberOfMaps; i++) {
      if (NULL != pMapArray[i].pTable) {
        dfFree(pMapArray[i].pTable);
        pMapArray[i].pTable = NULL;
      }
    }
    dfFree(pMapArray);
    pDevObj->system.dma.pMapArray = NULL;
  }
  return FALSE;
}

extern void
dfDmaMapperFree(
  DfDeviceObject* pDevObj)
{
  DfVxwDmaMapperContext* pMapArray;
  unsigned int i, numberOfMaps;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);

  pMapArray = pDevObj->system.dma.pMapArray;
  numberOfMaps = pDevObj->system.dma.numMap;
  if (NULL != pMapArray) {
    for (i = 0; i < numberOfMaps; i++) {
      if (NULL != pMapArray[i].pTable) {
        dfFree(pMapArray[i].pTable);
        pMapArray[i].pTable = NULL;
      }
    }
    dfFree(pMapArray);
    pDevObj->system.dma.pMapArray = NULL;
  } else {
    dfDebugPrint(0, ("*** dfDmaMapperFree: attempting to free mapper when mapper not allocated, pDevObj=0x%p\n", pDevObj));
  }
}

/*
** On the SVME182 under VxWorks 5.5, if we do cache invalidation after
** completing a DMA transfer in the direction device -> memory, the data in
** the buffer appears to be stale as if cache invalidation never happened.
**
** Workaround for now is to do invalidation (for device -> memory) as well as
** flushes (for memory -> device) during setup, and do nothing by way of
** teardown.
*/
#define CACHE_INVALIDATE_AFTER_DMA (0)

extern boolean_t
dfDmaMapperSetup(
  DfDeviceObject* pDevObj,
  unsigned int mapIndex,
  DfBufferDescription* pBufferDescription,
  DfDmaMapperCallback* pCallback,
  void* pCallbackContext,
  boolean_t bWriteToDevice,
  boolean_t bFixedLocalAddress,
  DfDmaMapperSlice* pSlices,
  size_t numSlice,
  DfDmaMapperPosition* pStartPosition)
{
  DfDmaMapperSlice* pSlice;
  DfDmaMapperNode* pNode;
  DfVxwDmaMapperContext* pMapperContext;
  DfDmaMapperPosition newPosition;
  size_t sliceIndex, slicePosition;
  size_t sliceOffset, sliceLength, sliceRemaining = 0;
  uint64_t sliceLocalAddress;
  size_t chunk, maxMapSize;
  unsigned int numEntry, maxEntry;
  uint64_t busAddress;
  uint8_t* pBufferAddress;
  uint8_t* pAddress;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);
  dfAssert(numSlice > 0);
  dfAssert(mapIndex < pDevObj->system.dma.numMap);

  maxMapSize = pDevObj->system.dma.maxMapSize;

  pBufferAddress = (uint8_t*)pBufferDescription->pAddress;

  sliceIndex = pStartPosition->sliceIndex;
  dfAssert(sliceIndex < numSlice);
  slicePosition = pStartPosition->slicePosition;

  pMapperContext = &pDevObj->system.dma.pMapArray[mapIndex];
  maxEntry = pDevObj->system.dma.maxMapRegister;
  pNode = pMapperContext->pTable;

  numEntry = 0;
  while (sliceIndex < numSlice && numEntry < maxEntry) {
    /* Get current offset into current slice, and remaining bytes in current slice */
    pSlice = pSlices + sliceIndex;
    pAddress = pBufferAddress + pSlice->offset + slicePosition;
    sliceOffset = pSlice->offset + slicePosition;
    sliceLength = pSlice->length;
    dfAssert(slicePosition < sliceLength);
    sliceRemaining = sliceLength - slicePosition;
    sliceLocalAddress = pSlice->localAddress;
    if (!bFixedLocalAddress) {
      sliceLocalAddress += slicePosition;
    }
    /* Calculate how many bytes we can transfer this time around */
    chunk = (sliceRemaining > maxMapSize) ? maxMapSize : sliceRemaining;
    if (!dfVxwLegacyCpuToBusAddress(pDevObj, pAddress, &busAddress)) {
      dfAssert(FALSE);
    }
#if CACHE_INVALIDATE_AFTER_DMA
    if (bWriteToDevice) {
      dfVxwLegacyCacheSync(pAddress, TRUE, chunk);
    }
#else
    dfVxwLegacyCacheSync(pAddress, bWriteToDevice, chunk);
#endif
    pNode->busAddress = busAddress;
    pNode->length = chunk;
    pNode->localAddress = sliceLocalAddress;
    pNode->system.pKernelAddress = pAddress;
    pNode++;
    /* Increment pointers etc. */
    slicePosition += chunk;
    if (slicePosition == sliceLength) {
      /* Reached end of current slice */
      sliceIndex++;
      slicePosition = 0;
    }
    numEntry++;
  }

  pMapperContext->curr.bWriteToDevice = bWriteToDevice;
  pMapperContext->curr.numEntry = numEntry;
  newPosition.sliceIndex = sliceIndex;
  newPosition.slicePosition = slicePosition;
  
  /* Finished, so call the client's callback function */
  pCallback(
    pDevObj,
    pMapperContext->index,
    pMapperContext->pTable,
    numEntry,
    &newPosition,
    pCallbackContext);

  return TRUE;
}

extern void
dfDmaMapperTeardown(
  DfDeviceObject* pDevObj,
  unsigned int mapIndex)
{
#if CACHE_INVALIDATE_AFTER_DMA
  DfDmaMapperNode* pNode;
  DfVxwDmaMapperContext* pMapperContext;
  unsigned int i, numEntry;
  boolean_t bWriteToDevice;

  dfAssert(pDevObj->hardwareId.busType == DfBusPci);

  pMapperContext = &pDevObj->system.dma.pMapArray[mapIndex];
  pNode = pMapperContext->pTable;
  numEntry = pMapperContext->curr.numEntry;
  bWriteToDevice = pMapperContext->curr.bWriteToDevice;

  if (!bWriteToDevice) {
    for (i = 0; i < numEntry; i++) {
      dfVxwLegacyCacheSync(pNode->system.pKernelAddress, FALSE, pNode->length);
      pNode++;
    }
  }
#else
  /* Do nothing */
#endif
}

#endif
