/*
** File: legacy_methods.c  
** Project: VxWorks Driver Framework
** Purpose: Legacy (non-VxBus) functions
**
** (C) Copyright Alpha Data 2010
**
** This file contains all platform-specific code for the ADB3 VxWorks legacy
** (non-VxBus) driver. The target platform is selected as compile-time by
** arranging for the appropriate preprocessor macro (e.g. MV5500) to be
** defined.
**
** The following functions are (in general) platform-specific in a non-VxBus
** driver:
**
**   o Translating bus addresses to CPU addresses, for register access
**   o Translating CPU addressses to bus address, for DMA transfers
**   o Data cache synchronization, for DMA transfers
**   o Interrupt hooking and unhooking
*/

#include "df.h"
#include "dfvxworks.h"

#if !defined(ADB3_VXBUS)
static void
_dfVxwIsrStub(
  DfInterruptObject* pInterruptObject)
{
  boolean_t bHandled;

  dfAssert(pInterruptObject->system.bConnected);
  bHandled = pInterruptObject->system.pIsr(pInterruptObject, pInterruptObject->system.pContext);
  /*
  ** There's currently no means of passing any information back to the VxWorks
  ** kernel about whether or not the interrupt was handled by us, so we just
  ** discard the value of 'bHandled'.
  */
}
#endif

#if !defined(ADB3_VXBUS)
boolean_t 
dfVxwLegacyBusToCpuAddress(
  DfDeviceObject* pDevObj,
  uint64_t busAddress,
  void** pKernelAddress)
{
  *pKernelAddress = (void*)(uintptr_t)busAddress;

  return TRUE;
}
#else
boolean_t 
dfVxwLegacyBusToCpuAddress(
  DfDeviceObject* pDevObj,
  uint64_t busAddress,
  void** pKernelAddress)
{
  dfAssert(FALSE);
  return FALSE;
}
#endif

#if !defined(ADB3_VXBUS_DMA)
boolean_t 
dfVxwLegacyCpuToBusAddress(
  DfDeviceObject* pDevObj,
  void* pKernelAddress,
  uint64_t* pBusAddress)
{
  *pBusAddress = (uint64_t)(uintptr_t)pKernelAddress;

  return TRUE;
}
#else
boolean_t 
dfVxwLegacyCpuToBusAddress(
  DfDeviceObject* pDevObj,
  void* pKernelAddress,
  uint64_t* pBusAddress)
{
  dfAssert(FALSE);
  return FALSE;
}
#endif

#if !defined(ADB3_VXBUS_DMA)
void
dfVxwLegacyCacheSync(
  void* pKernelAddress,
  boolean_t bWriteToDevice,
  size_t n)
{
  STATUS status;

# if defined(__i386) || defined(__i386__) || defined(__I386__) || defined(__386)

  /*
  ** x86 machine generally maintain coherency in hardware for DMA transfers, but
  ** call cacheLib anyway in case we are running on something that doesn't.
  */

  if (bWriteToDevice) {
    /* Data is being read from memory by a bus-master device */
    status = cacheFlush(DATA_CACHE, pKernelAddress, n);
  } else {
    /* Data is being deposited in memory by a bus-master device */
    status = cacheInvalidate(DATA_CACHE, pKernelAddress, n);
  }

# elif defined (__ppc) || defined(__PPC__)

  if (bWriteToDevice) {
    /* Data is being read from memory by a bus-master device */
    status = cacheFlush(DATA_CACHE, pKernelAddress, n);
  } else {
    /* Data is being deposited in memory by a bus-master device */
    status = cacheInvalidate(DATA_CACHE, pKernelAddress, n);
  }

# else

#   error "Unrecognized CPU architecture. Additional code is required in this file for your CPU architecture."
#   error "Supported CPU architectures: PPC, x86"

# endif

  dfAssert(status == OK);
}
#else
void
dfVxwLegacyCacheSync(
  void* pKernelAddress,
  boolean_t bWriteToDevice,
  size_t n)
{
  dfAssert(FALSE);
}
#endif

#if !defined(ADB3_VXBUS)
boolean_t
dfVxwLegacyInterruptConnect(
  DfDeviceObject* pDevObj,
  DfInterruptObject* pInterruptObject,
  DfBusResources* pBusResources,
  DfBusType busType,
  unsigned int interruptIndex,
  DfIsr* pIsr,
  void* pContext)
{
  DfPciIrqLine* pPciIrqLine;
  VOIDFUNCPTR* irqVector;
  int irqNumber;

  dfAssert(busType == DfBusPci);

  pPciIrqLine = &pBusResources->variant.pci.irq[interruptIndex];
  irqNumber = pPciIrqLine->variant.translated.intLine;
  irqVector = INUM_TO_IVEC(irqNumber);

# if defined(__i386) || defined(__i386__) || defined(__I386__) || defined(__386)

  /* Generic x86 method */
  if (pciIntConnect(irqVector, _dfVxwIsrStub, (int)pInterruptObject) != OK) {
    dfDebugPrint(0, ("*** dfVxwLegacyInterruptConnect(x86 generic): pciIntConnect() failed, irqVector=%p\n", (void*)irqVector));
    return FALSE;
  }
  if (sysIntEnablePIC(irqNumber) != OK) {
    dfDebugPrint(0, ("*** dfVxwLegacyInterruptConnect(x86 generic): sysIntEnablePIC() failed, irqNumber=%d(0x%x)\n", irqNumber, irqNumber));
    /* Undo the pciIntConnect before we error out */
    if (pciIntDisconnect2(irqVector, _dfVxwIsrStub, (int)pInterruptObject) != OK) {
      dfDebugPrint(0, ("*** dfVxwLegacyInterruptConnect(x86 generic): pciIntDisconnect2() failed, irqVector=%p\n", irqVector));
    }
    return FALSE;
  }

# elif defined (__ppc) || defined(__PPC__)

#   if defined(MV5500)

  /* MV5500-specific method */
  if (intConnect(irqVector, _dfVxwIsrStub, (int)pInterruptObject) != OK) {
    dfDebugPrint(0, ("*** dfVxwLegacyInterruptConnect(MV5500): intConnect() failed, irqVector=0x%p\n", (void*)irqVector));
    return FALSE;
  }
  if (intEnable(irqNumber) != OK) {
    dfDebugPrint(0, ("*** dfVxwLegacyInterruptConnect(MV5500): intEnable() failed, irqNumber=%d(0x%x)\n", irqNumber, irqNumber));
    /* intDisconnect() appears not to work on the MV5500, at least for PCI interrupts */
    return FALSE;
  }

#   else

  /* Generic PPC method */
  if (pciIntConnect(irqVector, _dfVxwIsrStub, (int)pInterruptObject) != OK) {
    dfDebugPrint(0, ("*** dfVxwLegacyInterruptConnect(PPCGeneric): pciIntConnect() failed, irqVector=0x%p\n", (void*)irqVector));
    return FALSE;
  }
  if (intEnable(irqNumber) != OK) {
    /* Undo the pciIntConnect before we error out */
    dfDebugPrint(0, ("*** dfVxwLegacyInterruptConnect(PPC generic): intEnable() failed, irqNumber=%d(0x%x)\n", irqNumber, irqNumber));
    if (pciIntDisconnect2(irqVector, _dfVxwIsrStub, (int)pInterruptObject) != OK) {
      dfDebugPrint(0, ("*** dfVxwLegacyInterruptConnect(PPC generic): pciIntDisconnect2() failed, irqVector=%p\n", irqVector));
    }
    return FALSE;
  }

#   endif

# else

#   error "Unrecognized processor architecture. Additional code is required in this file for your architecture."
#   error "Supported architectures: PPC, x86"

# endif

  /* Save information required by dfVxwLegacyInterruptDisconnect() */
  pInterruptObject->system.legacy.irqVector = irqVector;
  pInterruptObject->system.legacy.irqNumber = irqNumber;

  return TRUE;
}
#else
boolean_t
dfVxwLegacyInterruptConnect(
  DfDeviceObject* pDevObj,
  DfInterruptObject* pInterruptObject,
  DfBusResources* pBusResources,
  DfBusType busType,
  unsigned int interruptIndex,
  DfIsr* pIsr,
  void* pContext)
{
  dfAssert(FALSE);
  return FALSE;
}
#endif

#if !defined(ADB3_VXBUS)
void
dfVxwLegacyInterruptDisconnect(
  DfDeviceObject* pDevObj,
  DfInterruptObject* pInterruptObject)
{
  VOIDFUNCPTR* irqVector;
  int irqNumber;

  irqNumber = pInterruptObject->system.legacy.irqNumber;
  irqVector = pInterruptObject->system.legacy.irqVector;

# if defined(__i386) || defined(__i386__) || defined(__I386__) || defined(__386)

  /* Generic x86 method */
  if (sysIntDisablePIC(irqNumber) != OK) {
    dfDebugPrint(0, ("*** dfInterruptDisconnect(x86 generic): sysIntEnablePIC() failed, irqNumber=%d(0x%x)\n", irqNumber, irqNumber));
  }
  if (pciIntDisconnect2(irqVector, _dfVxwIsrStub, (int)pInterruptObject) != OK) {
    dfDebugPrint(0, ("*** dfInterruptDisconnect(x86 generic): pciIntDisconnect2() failed, irqVector=%p pInterruptObject=%p\n", irqVector, pInterruptObject));
  }

# elif defined (__ppc) || defined(__PPC__)

#   if defined(MV5500)

  /* MV5500-specific method */
  if (intDisable(irqNumber) != OK) {
    dfDebugPrint(0, ("*** dfInterruptDisconnect(MV5500): intDisable() failed, irqNumber=%d(0x%x)\n", irqNumber, irqNumber));
  }
  /* intDisconnect() appears not to work on the MV5500, at least for PCI interrupts */

#   else

  /* Generic PPC method */
  if (intDisable(irqNumber) != OK) {
    dfDebugPrint(0, ("*** dfInterruptDisconnect(PPC generic): intDisable() failed, irqNumber=%d(0x%x)\n", irqNumber, irqNumber));
  }
  if (pciIntDisconnect2(irqVector, _dfVxwIsrStub, (int)pInterruptObject) != OK) {
    dfDebugPrint(0, ("*** dfInterruptDisconnect(PPC generic): pciIntDisconnect2() failed, irqVector=%p pInterruptObject=%p\n", irqVector, pInterruptObject));
  }

#   endif

# else

#   error Unrecognized processor architecture. Additional code is required in this file for your architecture.
#   error Supported architectures: PPC, x86

# endif
}
#else
void
dfVxwLegacyInterruptDisconnect(
  DfDeviceObject* pDevObj,
  DfInterruptObject* pInterruptObject)
{
  dfAssert(FALSE);
}
#endif
