/*
** File: mmap.c  
** Project: VxWorks Driver Framework
** Purpose: Mechanism by which driver can register regions of device memory that are
**          eligible to be mapped into a user-mode process' address space.
**
** (C) Copyright Alpha Data 2010
*/

#include "df.h"
#include "dfvxworks.h"

DfMmappableRegion*
dfMmappableRegionFind(
	DfDeviceObject* pDevObj,
  unsigned int regionTag)
{
  DfMmappableRegion** ppRegion;
  DfMmappableRegion* pRegion;

  dfDebugPrint(3, ("dfMmappableRegionFind: entered, pDevObj=0x%p regionTag=%lu(0x%lx)\n", pDevObj, (unsigned long)regionTag, (unsigned long)regionTag));

  ppRegion = &pDevObj->system.pMmappableRegions;
  pRegion = pDevObj->system.pMmappableRegions;
  while (1) {
    pRegion = *ppRegion;
    if (NULL == pRegion) {
      break;
    }
    if (regionTag == pRegion->system.regionTag) {
      return pRegion;
    }
    ppRegion = &pRegion->system.pNext;
  }

  return NULL;
}

DfMmappableStatus
dfMmappableRegionRegister(
	DfDeviceObject* pDevObj,
  uint32_t regionTag,
  void* pKernelBase,
  size_t kernelSize,
  uint64_t translatedBase,
  uint64_t translatedSize)
{
  DfMmappableRegion* pRegion;

  dfDebugPrint(3, ("dfMmappableRegionRegister: entered, pDevObj=0x%p regionTag=%lu(0x%lx) pKernelBase=0x%p kernelSize=0x%p translatedBase=0x%08lx_%08lx translatedSize=0x%08lx_%08lx\n",
    pDevObj, (unsigned long)regionTag, (unsigned long)regionTag, pKernelBase, (void*)(uintptr_t)kernelSize, dfSplitUint64(translatedBase), dfSplitUint64(translatedSize)));

  /* Sanity check */
  pRegion = dfMmappableRegionFind(pDevObj, regionTag);
  dfAssert(NULL == pRegion);

  pRegion = dfMalloc(sizeof(DfMmappableRegion));
  if (NULL == pRegion) {
    return DfMmappableNoMemory;
  }
  pRegion->system.regionTag = regionTag;
  pRegion->system.pKernelBase = pKernelBase;
  pRegion->system.kernelSize = kernelSize;
  pRegion->system.pNext = pDevObj->system.pMmappableRegions;
  pDevObj->system.pMmappableRegions = pRegion;
  return DfMmappableSuccess;
}

DfMmappableStatus
dfMmappableRegionUnregister(
	DfDeviceObject* pDevObj,
  uint32_t regionTag)
{
  DfMmappableRegion** ppRegion;
  DfMmappableRegion* pRegion;

  dfDebugPrint(3, ("dfMmappableRegionUnregister: entered, pDevObj=0x%p regionTag=%lu(0x%lx)\n", pDevObj, (unsigned long)regionTag, (unsigned long)regionTag));

  ppRegion = &pDevObj->system.pMmappableRegions;
  pRegion = pDevObj->system.pMmappableRegions;
  while (1) {
    pRegion = *ppRegion;
    if (NULL == pRegion) {
      break;
    }
    if (regionTag == pRegion->system.regionTag) {
      *ppRegion = pRegion->system.pNext;
      dfFree(pRegion);
      return DfMmappableSuccess;
    } else {
      ppRegion = &pRegion->system.pNext;
    }
  }

  return DfMmappableInvalidRegion;
}

DfMmappableStatus
dfVxwMmapRegion(
	DfDeviceObject* pDevObj,
	DfClientObject* pClObj,
  uint32_t regionTag,
  uint64_t offset,
  uint64_t size,
  void** ppUserVirtualBase)
{
  DfSpinLockFlags f;
  void* pUserAddr = NULL;
  uint8_t* pKernelAddress = NULL;
  DfMmappedRegion* pRegion = NULL;
  DfMmappableRegion* pMappable = NULL;
  DfMmappableStatus status = DfMmappableSuccess;

  dfDebugPrint(3, ("dfVxwMmapRegion: entered, pDevObj=0x%p pClObj=0x%p regionTag=%lu(0x%lx) offset=0x%08lx_%08lx size=0x%08lx_%08lx\n",
    pDevObj, pClObj, (unsigned long)regionTag, (unsigned long)regionTag, dfSplitUint64(offset), dfSplitUint64(size)));

  /* Catch nonsense values */
  if (size == 0 || size > (size_t)-1 || offset + size < offset) {
    status = DfMmappableInvalidRegion;
    goto done;
  }

  /* Find the region and check bounds */
  pMappable = dfMmappableRegionFind(pDevObj, regionTag);
  if (NULL == pMappable) {
    status = DfMmappableInvalidRegion;
    goto done;
  }
  if (offset >= pMappable->system.kernelSize || offset + size > pMappable->system.kernelSize) {
    status = DfMmappableInvalidRegion;
    goto done;
  }

  /* Now do the mapping */
  pKernelAddress = (UINT8*)pMappable->system.pKernelBase;
  pRegion = dfMalloc(sizeof(DfMmappedRegion));
  if (NULL == pRegion) {
    status = DfMmappableNoMemory;
    goto done;
  }
  pUserAddr = pKernelAddress + offset;
  pRegion->system.pKernelBase = pKernelAddress;
  pRegion->system.kernelSize = size;
  pRegion->system.pUserVirtualBase = pUserAddr;
  *ppUserVirtualBase = pUserAddr;
  f = dfSpinLockGet(&pClObj->system.lock);
  pRegion->system.pNext = pClObj->system.pMmappedRegions;
  pClObj->system.pMmappedRegions = pRegion;
  dfSpinLockPut(&pClObj->system.lock, f);

done:
  if (DfMmappableSuccess != status) {
    if (NULL != pRegion) {
      dfFree(pRegion);
    }
  }
  return status;
}

DfMmappableStatus
dfVxwMunmapRegion(
	DfDeviceObject* pDevObj,
	DfClientObject* pClObj,
  void* pUserVirtualBase)
{
  DfMmappedRegion* pRegion;
  DfMmappedRegion** ppRegion;
  DfSpinLockFlags f;

  dfDebugPrint(3, ("dfVxwMunmapRegion: entered, pDevObj=0x%p pClObj=0x%p pUserVirtualBase=0x%p\n",
    pDevObj, pClObj, pUserVirtualBase));

  f = dfSpinLockGet(&pClObj->system.lock);
  ppRegion = &pClObj->system.pMmappedRegions;
  while (1) {
    pRegion = *ppRegion;
    if (NULL == pRegion) {
      dfSpinLockPut(&pClObj->system.lock, f);
      dfDebugPrint(1, ("dfVxwMunmapRegion: region 0x%p not found\n", pUserVirtualBase));
      return DfMmappableInvalidRegion;
    }
    if (pUserVirtualBase == pRegion->system.pUserVirtualBase) {
      *ppRegion = pRegion->system.pNext;
      break;
    }
    ppRegion = &pRegion->system.pNext;
  }
  dfSpinLockPut(&pClObj->system.lock, f);

  dfDebugPrint(3, ("dfVxwMunmapRegion: region found, pRegion=0x%p pKernelBase=0x%p\n",
    pRegion, pRegion->system.pKernelBase));

  dfFree(pRegion);

  return DfMmappableSuccess;
}

DfMmappableStatus
dfVxwMunmapAllRegions(
	DfDeviceObject* pDevObj,
	DfClientObject* pClObj)
{
  DfMmappedRegion* pRegion;
  DfSpinLockFlags f;

  while (1) {
    f = dfSpinLockGet(&pClObj->system.lock);
    pRegion = pClObj->system.pMmappedRegions;
    if (NULL == pRegion) {
      dfSpinLockPut(&pClObj->system.lock, f);
      break;
    } else {
      pClObj->system.pMmappedRegions = pRegion->system.pNext;
      dfSpinLockPut(&pClObj->system.lock, f);

      dfDebugPrint(3, ("dfVxwMunmapAllRegions: region found, pRegion=0x%p pKernelBase=0x%p\n",
        pRegion, pRegion->system.pKernelBase));

      dfFree(pRegion);
    }
  }

  return DfMmappableSuccess;
}
