/*
** File: userif.c  
** Project: VxWorks Driver Framework
** Purpose: Implements mechanism by which OS-independent driver code can register
**          an interface accessible by user-mode programs.
**
** (C) Copyright Alpha Data 2010
*/

#include "df.h"
#include "dfvxworks.h"
#include <ioctl_dfvxw.h>

/* Client object state, used to wait for outstanding requests to be finished when closing */
typedef enum _ClientState {
  ClientStateOpen = 0,    /* Not being closed */
  ClientStateClosing = 1, /* Being closed, before testing if there are outstanding client requests */
  ClientStateWaiting = 2  /* Waiting for outstanding client requests to be completed */
} ClientState;

typedef enum _NonBlockState {
  NonBlockStateIdle = 0,     /* No non-blocking operations in progress */
  NonBlockStateIoctl = 1,    /* Executing IOCTL handler */
  NonBlockStateDeferred = 2  /* Waiting on completion */
} NonBlockState;

/*
** ---------------------------------------------------------------------------
** Functions for managing client request lists
** ---------------------------------------------------------------------------
*/

DF_DECLARE_INLINE_FUNC(void, requestInit)(
  DfClientRequest* pReq,
  DfClientObject* pClObj)
{
  pReq->system.pClObj = pClObj;
  pReq->system.pCancelCallback = NULL;
  pReq->system.bCancel = FALSE;
  pReq->system.bCompleted = FALSE;
  pReq->system.pContext = NULL;
  pReq->system.pWaitEvent = NULL;
}

static DfIoStatus
deliverRequestBlocking(
  DfDeviceObject* pDevObj,
  DfClientObject* pClObj,
  DfOnIoctlMethod* pHandler,
  unsigned int reqCode,
  void* pIoctlStruct,
  unsigned int sizeIn,
  unsigned int sizeOut)
{
  DfClientRequest* pReq;
  DfIoStatus status;
  DfEvent ev;

  pReq = dfPoolAlloc(&pDevObj->system.request.pool, DfClientRequest);
  if (NULL == pReq) {
    return ENOMEM;
  }
  requestInit(pReq, pClObj);

  dfEventInit(&ev);
  pReq->system.pWaitEvent = &ev;
  status = pHandler(pDevObj, pClObj, pReq, reqCode, pIoctlStruct, sizeIn, sizeOut);
  if (status == DfIoStatusDeferred) {
    dfEventWait(pReq->system.pWaitEvent);
    status = pReq->system.status;
  }
  dfEventUninit(&ev);
  dfPoolFree(&pDevObj->system.request.pool, pReq);

  return status;
}

static DfIoStatus
deliverRequestNonBlocking(
  DfDeviceObject* pDevObj,
  DfClientObject* pClObj,
  DfOnIoctlMethod* pHandler,
  unsigned int reqCode,
  void* pIoctlStruct,
  unsigned int sizeIn,
  unsigned int sizeOut)
{
  DfIoStatus status;
  DfVxwClientRequestNonBlock* pReq;
  DfSpinLockFlags f;

  pReq = dfPoolAlloc(&pDevObj->system.request.poolNonBlock, DfVxwClientRequestNonBlock);
  if (NULL == pReq) {
    return ENOMEM;
  }
  requestInit(&pReq->clientRequest, pClObj);
  dfCopyUK(&pReq->ioctlStruct, pIoctlStruct, sizeIn);

  f = dfSpinLockGet(&pClObj->system.lock);
  if (NonBlockStateIdle != pClObj->system.nonblock.state) {
    dfSpinLockPut(&pClObj->system.lock, f);    
    dfPoolFree(&pDevObj->system.request.poolNonBlock, pReq);
    return EBUSY;
  } else {
    pClObj->system.nonblock.state = NonBlockStateIoctl;
    pClObj->system.nonblock.pHead = pReq;
    dfSpinLockPut(&pClObj->system.lock, f);
    status = pHandler(pDevObj, pClObj, &pReq->clientRequest, reqCode, &pReq->ioctlStruct, sizeIn, sizeOut);
    f = dfSpinLockGet(&pClObj->system.lock);
    if (DfIoStatusIsError(status)) {
      pClObj->system.nonblock.state = NonBlockStateIdle;
      pClObj->system.nonblock.pHead = NULL;
    } else {
      pClObj->system.nonblock.state = NonBlockStateDeferred;
    }
    dfSpinLockPut(&pClObj->system.lock, f);
    if (DfIoStatusIsError(status)) {
      dfPoolFree(&pDevObj->system.request.poolNonBlock, pReq);
    }
  }
  return status;
}

boolean_t
dfRequestClearCallback(
  DfClientRequest* pClientRequest)
{
  DfClientObject* pClObj;
  DfSpinLockFlags f;

  dfAssert(!pClientRequest->system.bCompleted);

  pClObj = pClientRequest->system.pClObj;

  f = dfSpinLockGet(&pClObj->system.lock);
  if (pClientRequest->system.bCancel) {
    dfSpinLockPut(&pClObj->system.lock, f);
    return FALSE;
  } else {
    dfListRemove(&pClientRequest->system.node);
    pClientRequest->system.pCancelCallback = NULL;
    dfSpinLockPut(&pClObj->system.lock, f);
    return TRUE;
  }
}

void
dfRequestComplete(
  DfClientRequest* pClientRequest,
  DfIoStatus status)
{
  DfClientObject* pClObj;
  DfSpinLockFlags f;

  dfAssert(!pClientRequest->system.bCompleted);
  dfAssert(pClientRequest->system.pCancelCallback == NULL);

  pClObj = pClientRequest->system.pClObj;

  f = dfSpinLockGet(&pClObj->system.lock);
  pClientRequest->system.status = status;
  pClientRequest->system.bCompleted = TRUE;
  if (NULL != pClientRequest->system.pWaitEvent) {
    /* Blocking */
    dfSpinLockPut(&pClObj->system.lock, f);
    dfEventSignal(pClientRequest->system.pWaitEvent);
  } else {
    /* Non-blocking */
    dfSpinLockPut(&pClObj->system.lock, f);
    selWakeupAll(&pClObj->system.nonblock.selWakeupList, SELREAD);
  }
}

boolean_t
dfRequestSetCallback(
  DfClientRequest* pClientRequest,
  DfCancelCallback* pCallback)
{
  DfClientObject* pClObj;
  DfSpinLockFlags f;

  dfAssert(!pClientRequest->system.bCompleted);
  dfAssert(pClientRequest->system.pCancelCallback == NULL);

  pClObj = pClientRequest->system.pClObj;

  f = dfSpinLockGet(&pClObj->system.lock);
  if (pClientRequest->system.bCancel) {
    dfSpinLockPut(&pClObj->system.lock, f);
    return FALSE;
  } else {
    pClientRequest->system.pCancelCallback = pCallback;
    dfListAddToTail(&pClObj->system.requests.list, &pClientRequest->system.node);
    dfSpinLockPut(&pClObj->system.lock, f);
    return TRUE;
  }
}

/*
** ---------------------------------------------------------------------------
** VxWorks driver framework special IOCTL handlers
** ---------------------------------------------------------------------------
*/

/* Request to map a region into user-mode application's address space */
static STATUS 
ioctlMmap(
	DfDeviceObject* pDevObj,
  DfClientObject* pClObj,
  void* pIoctlStruct,
  unsigned int inLength,
  unsigned int outLength)
{
  DfVxwIoctlMmap* pMapIoctlStruct = (DfVxwIoctlMmap*)pIoctlStruct;
  DfMmappableStatus mmapStatus;
  void* pUserVirtualBase;

  if (NULL == pIoctlStruct || inLength != sizeof(pMapIoctlStruct->in) || outLength != sizeof(pMapIoctlStruct->out)) {
    errno = EFAULT;
    return ERROR;
  }

  mmapStatus = dfVxwMmapRegion(pDevObj, pClObj, pMapIoctlStruct->in.regionTag, pMapIoctlStruct->in.offset, pMapIoctlStruct->in.length, &pUserVirtualBase); 
  switch (mmapStatus) {
  case DfMmappableSuccess:
    pMapIoctlStruct->out.pVirtualBase = pUserVirtualBase;
    return (STATUS)DfVxwMmapSuccess;

  case DfMmappableInvalidRegion:
    return (STATUS)DfVxwMmapInvalidRegion;

  case DfMmappableRegionTooLarge:
    return (STATUS)DfVxwMmapRegionTooLarge;

  default:
    return (STATUS)DfVxwMmapGeneralFailure;
  }
}

  /* Request to unmap a region from user-mode application's address space */
static STATUS
ioctlMunmap(
	DfDeviceObject* pDevObj,
  DfClientObject* pClObj,
  void* pIoctlStruct,
  unsigned int inLength)
{
  DfVxwIoctlMunmap* pUnmapIoctlStruct = (DfVxwIoctlMunmap*)pIoctlStruct;
  DfMmappableStatus mmapStatus;

  if (NULL == pIoctlStruct || inLength != sizeof(pUnmapIoctlStruct->in)) {
    errno = EFAULT;
    return ERROR;
  }

  mmapStatus = dfVxwMunmapRegion(pDevObj, pClObj, pUnmapIoctlStruct->in.pVirtualBase);
  switch (mmapStatus) {
  case DfMmappableSuccess:
    return (STATUS)DfVxwMunmapSuccess;

  case DfMmappableInvalidRegion:
    return (STATUS)DfVxwMunmapInvalidRegion;

  default:
    return (STATUS)DfVxwMunmapGeneralFailure;
  }
}

static STATUS
ioctlCancelRequests(
  DfDeviceObject* pDevObj,
  DfClientObject* pClObj)
{
  DfList complete;
  DfListNode* pHead;
  DfClientRequest* pCurrent;
  DfCancelCallback* pCancelCallback;
  DfSpinLockFlags f;

  dfDebugPrint(5, ("ioctlCancelRequests: entered\n"));

  /*
  ** Confiscate all current client requests but keep the list intact. Clears the
  ** list head pointer in the client object.
  */
  f = dfSpinLockGet(&pClObj->system.lock);
  dfListConfiscate(&complete, &pClObj->system.requests.list);
  dfSpinLockPut(&pClObj->system.lock, f);

  /* Now iterate through the confiscated list and cancel the client requests */
  while (1) {
    f = dfSpinLockGet(&pClObj->system.lock);
    pHead = dfListGetHead(&complete);
    if (pHead == NULL) {
      dfSpinLockPut(&pClObj->system.lock, f);
      break;
    }
    dfListRemove(pHead);
    pCurrent = DF_CONTAINER_OF(pHead, DfClientRequest, system.node);
    pCurrent->system.bCancel = TRUE;
    pCancelCallback = pCurrent->system.pCancelCallback;
    pCurrent->system.pCancelCallback = NULL;
    dfSpinLockPut(&pClObj->system.lock, f);      

    dfDebugPrint(2, ("ioctlCancelRequests: cancelling request 0x%p\n", pCurrent));
    dfAssert(NULL != pCancelCallback);
    pCancelCallback(pCurrent, DfIoStatusCancelled);
  }

  return OK;
}

static STATUS
ioctlFinishNonBlock(
  DfDeviceObject* pDevObj,
  DfClientObject* pClObj,
  void* pIoctlOut,
  unsigned int sizeOut)
{
  DfIoStatus status = OK;
  DfVxwClientRequestNonBlock* pReq;
  DfSpinLockFlags f;

  if (sizeOut > 0 && NULL == pIoctlOut) {
    errno = EFAULT;
    return ERROR;
  }

  f = dfSpinLockGet(&pClObj->system.lock);
  switch (pClObj->system.nonblock.state) {
  case NonBlockStateDeferred:
    pReq = pClObj->system.nonblock.pHead;
    dfAssert(NULL != pReq);
    if (pReq->clientRequest.system.bCompleted) {
      pClObj->system.nonblock.state = NonBlockStateIdle;
      pClObj->system.nonblock.pHead = NULL;
      dfSpinLockPut(&pClObj->system.lock, f);
      status = pReq->clientRequest.system.status;
      if (status == DfIoStatusSuccess && 0 != sizeOut) {
        dfCopyKU(pIoctlOut, &pReq->ioctlStruct, sizeOut);
      }
      dfPoolFree(&pDevObj->system.request.poolNonBlock, pReq);
      if (status == DfIoStatusSuccess) {
        return OK;
      } else {
        if (DfIoStatusIsUser(status)) {
          return (STATUS)status;
        } else {
          errno = (int)status;
          return ERROR;
        }
      }
    } else {
      dfSpinLockPut(&pClObj->system.lock, f);
      errno = EINPROGRESS;
      return ERROR;
    }
    break;

  case NonBlockStateIdle:
  case NonBlockStateIoctl:
  default:
    dfSpinLockPut(&pClObj->system.lock, f);
    errno = EPIPE;
    return ERROR;
  }
}

static int
specialIoctl(
  DfDeviceObject* pDevObj,
  DfClientObject* pClObj,
  unsigned int code,
  void* pIoctl,
  unsigned int sizeIn,
  unsigned int sizeOut)
{
  dfDebugPrint(5, ("specialIoctl: entered, code=%u(0x%x) sizeIn=%u(0x%x) sizeOut=%u(0x%x) pIoctl=%p\n",
    code, code, sizeIn, sizeIn, sizeOut, sizeOut, pIoctl));

  switch (code) {
  case DFVXW_IOCTLCODE_MMAP:
    return ioctlMmap(pDevObj, pClObj, pIoctl, sizeIn, sizeOut);

  case DFVXW_IOCTLCODE_MUNMAP:
    return ioctlMunmap(pDevObj, pClObj, pIoctl, sizeIn);

  case DFVXW_IOCTLCODE_CANCEL:
    return ioctlCancelRequests(pDevObj, pClObj);

  case DFVXW_IOCTLCODE_FINISH:
    return ioctlFinishNonBlock(pDevObj, pClObj, pIoctl, sizeOut);

  default:
    errno = ENOSYS;
    return ERROR;
  }
}

static int
ioctlSelect(
  DfClientObject* pClObj,
  SEL_WAKEUP_NODE* pNode)
{
  DfVxwClientRequestNonBlock* pReq;
  DfSpinLockFlags f;
  boolean_t bReady;

  if (selWakeupType(pNode) != SELREAD) {
    return OK;
  }

  f = dfSpinLockGet(&pClObj->system.lock);
  switch (pClObj->system.nonblock.state) {
  case NonBlockStateIoctl:
  case NonBlockStateDeferred:
    pReq = pClObj->system.nonblock.pHead;
    dfAssert(NULL != pReq);
    if (pReq->clientRequest.system.bCompleted) {
      bReady = TRUE;
    } else { 
      bReady = FALSE;
    }
    break;

  case NonBlockStateIdle:
    bReady = FALSE;
    break;

  default:
    dfAssert(FALSE);
    bReady = FALSE;
    break;
  }
  if (!bReady) {
    selNodeAdd(&pClObj->system.nonblock.selWakeupList, pNode);
  }
  dfSpinLockPut(&pClObj->system.lock, f);

  if (bReady) {
    selWakeup(pNode);
  }
  return OK;
}

static int
ioctlUnselect(
  DfClientObject* pClObj,
  SEL_WAKEUP_NODE* pNode)
{
  selNodeDelete(&pClObj->system.nonblock.selWakeupList, pNode);
  return OK;
}

/*
** ---------------------------------------------------------------------------
** file_operations methods
**
** These functions are called for every DfInterface object, regardless of
** class. In general, after identifying the DfInterface object associated with
** the 'DfClientObject' object, they call the methods associated with the
** DfInterface object in order to achieve the desired behavior.
** ---------------------------------------------------------------------------
*/

static int
fopsOpen( 
  DfInterface* pInterface,
  char* pName,
  int mode)
{
  DfDeviceObject* pDevObj;
  DfClientObject* pClObj;
  DfOnOpenMethod* pOnOpen;
  DfOpenStatus status;

  dfDebugPrint(3, ("fopsOpen: entered, pInterface=0x%p\n", pInterface));

  dfAssert(NULL != pInterface);
  pDevObj = pInterface->system.pDevObj;

  pClObj = (DfClientObject*)dfMalloc(sizeof(*pClObj));
  if (pClObj == NULL) {
    errno = ENOMEM;
    return ERROR;
  }
  dfSpinLockInit(&pClObj->system.lock);
  pClObj->system.pDevObj = pDevObj;
  pClObj->system.pInterface = pInterface;
  pClObj->system.pMmappedRegions = NULL;
  pClObj->system.cleanup.state = ClientStateOpen;
  pClObj->system.requests.count = 0;
  dfListInit(&pClObj->system.requests.list);
  pClObj->system.nonblock.state = NonBlockStateIdle;
  pClObj->system.nonblock.pHead = NULL;
  /* initialize wakeup list */
  selWakeupListInit(&pClObj->system.nonblock.selWakeupList);
  pClObj->system.pContext = NULL;

  pOnOpen = pInterface->system.methods.pOnOpen;
  if (NULL != pOnOpen) {
    status = pOnOpen(pDevObj, pClObj, pInterface, TRUE /* No access control in VxWorks */);
    if (status) {
      dfFree(pClObj);
      errno = (int)status;
      return ERROR;
    }
  }

  return (int)pClObj; /* Success */
}

static int
fopsClose(
  DfClientObject* pClObj)
{
  DfDeviceObject* pDevObj;
  DfInterface* pInterface;
  DfOnCleanupMethod* pOnCleanup;
  DfOnCloseMethod* pOnClose;
  DfSpinLockFlags f;
  DfVxwClientRequestNonBlock* pReq = NULL; /* init to prevent compiler whine */
  DfCancelCallback* pCancelCallback = NULL;
  boolean_t bMustWait = FALSE;
  
  dfDebugPrint(3, ("fopsClose: entered, pClObj=0x%p\n", pClObj));

  dfAssert(NULL != pClObj);
  pInterface = pClObj->system.pInterface;
  pDevObj = pClObj->system.pDevObj;

  f = dfSpinLockGet(&pClObj->system.lock);
  if (pClObj->system.nonblock.state != NonBlockStateIdle) {
    dfAssert(pClObj->system.nonblock.state >= NonBlockStateDeferred);
    pReq = pClObj->system.nonblock.pHead;
    dfAssert(NULL != pReq);
    if (!pReq->clientRequest.system.bCompleted) {
      dfAssert(!pReq->clientRequest.system.bCancel);
      pReq->clientRequest.system.bCancel = TRUE;
      pCancelCallback = pReq->clientRequest.system.pCancelCallback;
      if (NULL != pCancelCallback) {
        pReq->clientRequest.system.pCancelCallback = NULL;
        dfListRemove(&pReq->clientRequest.system.node);
      }
      bMustWait = TRUE;
    }
  }
  dfSpinLockPut(&pClObj->system.lock, f);
  if (NULL != pCancelCallback) {
    pCancelCallback(&pReq->clientRequest, DfIoStatusCancelled);
  }
  
  pOnCleanup = pInterface->system.methods.pOnCleanup;
  if (NULL != pOnCleanup) {
    pOnCleanup(pDevObj, pClObj);
  }

  if (bMustWait) {
    dfEventWait(pReq->clientRequest.system.pWaitEvent);
    dfPoolFree(&pDevObj->system.request.poolNonBlock, pReq);
    pClObj->system.nonblock.pHead = NULL;
    pClObj->system.nonblock.state = NonBlockStateIdle;
  }

  selWakeupListTerm(&pClObj->system.nonblock.selWakeupList);

  /* Clean up mmap'ed regions of user address space */
  dfVxwMunmapAllRegions(pDevObj, pClObj);

  pOnClose = pInterface->system.methods.pOnClose;
  if (NULL != pOnClose) {
    pOnClose(pDevObj, pClObj);
  }

  /* There should be no regions still mapped */
  dfAssert(pClObj->system.pMmappedRegions == NULL);

  dfFree(pClObj);

  return OK;
}

static int
fopsIoctl(
  DfClientObject* pClObj,
  int code,
  int arg)
{
  DfDeviceObject* pDevObj;
  DfInterface* pInterface;
  DfOnIoctlMethod* pHandler;
  DfIoStatus status = DfIoStatusInvalidCode;
  unsigned int sizeIn, sizeOut, reqCode;
  boolean_t bNonBlock;

  dfDebugPrint(5, ("fopsIoctl: entered, pClObj=0x%p code=%u(0x%x) arg=%u(0x%x)\n",
		   pClObj, code, code, arg, arg));

  dfAssert(NULL != pClObj);
  pDevObj = pClObj->system.pDevObj;
  pInterface = pClObj->system.pInterface;

  if ((((unsigned int)code >> 28) & 0xFU) != 0x3U) {
    switch (code) {
    case FIOSELECT:
      return ioctlSelect(pClObj, (SEL_WAKEUP_NODE*)(uintptr_t)arg);

    case FIOUNSELECT:
      return ioctlUnselect(pClObj, (SEL_WAKEUP_NODE*)(uintptr_t)arg);

    default:
      errno = ENOSYS;
      return ERROR;
    }
  }

  /* The IOCTL structure in and out sizes are encoded in the IOC size field of 'code' */
  sizeIn = DFVXW_GET_IOCTLINSIZE(code);
  sizeOut = DFVXW_GET_IOCTLOUTSIZE(code);
  reqCode = DFVXW_GET_IOCTLCODE(code);
  if (reqCode & DFVXW_IOCTLCODE_SPECIAL) {
    return specialIoctl(pDevObj, pClObj, reqCode, (void*)(size_t)arg, sizeIn, sizeOut);
  }
  bNonBlock = (reqCode & DFVXW_IOCTLCODE_NONBLOCK) ? TRUE : FALSE;
  reqCode &= 0xffU;
  dfAssert(!bNonBlock || 0 == sizeOut);
  if ((sizeIn != 0 || sizeOut != 0) && 0 == arg) {
    errno = EFAULT;
    return ERROR;
  }

  dfDebugPrint(5, ("fopsIoctl: sizeIn=%u sizeOut=%u reqCode=%u bNonBlock=%s\n",
    sizeIn, sizeOut, reqCode, bNonBlock ? "TRUE" : "FALSE"));

  pHandler = pInterface->system.methods.pOnIoctl;
  if (NULL != pHandler) {
    if (bNonBlock) {
      status = deliverRequestNonBlocking(pDevObj, pClObj, pHandler, reqCode, (void*)(uintptr_t)arg, sizeIn, sizeOut);
    } else {
      status = deliverRequestBlocking(pDevObj, pClObj, pHandler, reqCode, (void*)(uintptr_t)arg, sizeIn, sizeOut);
    }
  }

  if (status == DfIoStatusSuccess) {
    return OK;
  } else {
    if (DfIoStatusIsUser(status)) {
      return (STATUS)status;
    } else {
      errno = (int)status;
      return ERROR;
    }
  }
}

/*
** -----------------------------------------------------------------
** Functions exported by VxWorks driver framework
** -----------------------------------------------------------------
*/

boolean_t
dfInterfaceClassRegister(
  DfDriverObject* pDrvObj,
  const char* pClassName,
  DfInterfaceClass** ppClass)
{
  DfInterfaceClass* pClass;
  int result;

  pClass = (DfInterfaceClass*)dfMalloc(sizeof(*pClass));
  if (NULL == pClass) {
    dfDebugPrint(0, ("*** dfInterfaceClassRegister: failed to allocate struct, pDrvObj=0x%p\n", pDrvObj));
    return FALSE;
  }
  dfStrCpy(pClass->system.name, DF_ARRAY_LENGTH(pClass->system.name), pClassName);
  pClass->system.numInterface = 0;
  result = iosDrvInstall(
    (FUNCPTR)NULL,
    (FUNCPTR)NULL,
    fopsOpen,
    fopsClose,
    (FUNCPTR)NULL,
    (FUNCPTR)NULL,
    fopsIoctl);
  if (ERROR == result) {
    dfDebugPrint(0, ("*** dfInterfaceClassRegister: failed to install driver, pDrvObj=0x%p\n", pDrvObj));
    dfFree(pClass);
    return FALSE;
  }
  pClass->system.drvNum = result;

  *ppClass = pClass;
  return TRUE;
}

void
dfInterfaceClassUnregister(
  DfInterfaceClass* pClass)
{
  dfAssert(NULL != pClass);
  dfAssert(NULL != pClass && 0 == pClass->system.numInterface);
  if (OK != iosDrvRemove(pClass->system.drvNum, FALSE)) {
    dfDebugPrint(0, ("*** dfInterfaceClassUnregister: failed to uninstall driver, drvNum=%d\n", pClass->system.drvNum));
  }
  dfFree(pClass);
}

DF_ATTR_THREAD_CONTEXT
boolean_t
dfInterfaceRegister(
  DfDeviceObject* pDevObj,
  DfInterface* pInterface,
  DfInterfaceClass* pClass,
  const DfInterfaceMethods* pMethods)
{
  unsigned int index;

  index = pClass->system.numInterface;
  pInterface->system.methods.pOnCleanup = pMethods->pOnCleanup;
  pInterface->system.methods.pOnClose = pMethods->pOnClose;
  pInterface->system.methods.pOnIoctl = pMethods->pOnIoctl;
#if DF_NEED_THUNK
  pInterface->system.methods.pOnIoctlThunk = pMethods->pOnIoctlThunk;
#endif
  pInterface->system.methods.pOnOpen = pMethods->pOnOpen;
  pInterface->system.pDevObj = pDevObj;
  pInterface->system.pNext = NULL;
  pInterface->system.pContext = NULL;
  pInterface->system.bEnabled = FALSE;
  pDevObj->system.interfaceList.pHead = pInterface;
  pInterface->system.pClass = pClass;
  pInterface->system.index = index;
#if (_WRS_VXWORKS_MAJOR >= 6)
  snprintf(pInterface->system.name, DF_ARRAY_LENGTH(pClass->system.name), "/%s/%u", pClass->system.name, index);
#else
  /* Need VxWorks 6.x for snprintf() */
  sprintf(pInterface->system.name, "/%s/%u", pClass->system.name, index);
#endif
  pClass->system.numInterface = index + 1;
  return TRUE;
}

DF_ATTR_THREAD_CONTEXT
boolean_t
dfInterfaceUnregister(
  DfDeviceObject* pDevObj,
  DfInterface* pInterface)
{
  DfInterfaceClass* pClass;
  DfInterface** pp;
  DfInterface* p;

  dfAssert(!pInterface->system.bEnabled);
  dfAssert(NULL != pInterface->system.pClass);

  pClass = pInterface->system.pClass;
  pClass->system.numInterface--;

  pp = &pDevObj->system.interfaceList.pHead;
  while (1) {
    p = *pp;
    if (NULL == p) {
      break;
    }
    if (p == pInterface) {
      *pp = p->system.pNext;
      break;
    }
  }

  return (NULL == p) ? FALSE : TRUE;
}

DF_ATTR_THREAD_CONTEXT
boolean_t
dfInterfaceEnable(
  DfDeviceObject* pDevObj,
  DfInterface* pInterface,
  boolean_t bEnable)
{
  DfInterfaceClass* pClass;

  pClass = pInterface->system.pClass;
  if (bEnable) {
    dfAssert(!pInterface->system.bEnabled);
    if (OK != iosDevAdd(&pInterface->system.iosDev, pInterface->system.name, pClass->system.drvNum)) {
      dfDebugPrint(0, ("*** dfInterfaceEnable: failed to add device, pDevObj=0x%p\n", pDevObj));
      return FALSE;
    }
    pInterface->system.bEnabled = TRUE;
    return TRUE;
  } else {
    dfAssert(pInterface->system.bEnabled);
    pInterface->system.bEnabled = FALSE;
#if (_WRS_VXWORKS_MAJOR >= 6)
    if (OK != iosDevDelete(&pInterface->system.iosDev)) {
      dfDebugPrint(0, ("+++ dfInterfaceEnable: failed to delete device, pDevObj=0x%p\n", pDevObj));
      return FALSE;
    }
#else
    /* iosDevDelete() returns void prior to VxWorks 6.x */
    iosDevDelete(&pInterface->system.iosDev);
#endif
    return TRUE;
  }
}
