/*
** File: dma.c  
** Project: ADMXRC3 module driver
** Purpose: OS-independent IOCTL handlers for DMA transfers.
**
** (C) Copyright Alpha Data 2009
*/

#include <df.h>
#include "device.h"
#include "dma.h"
#include "lock.h"
#include <ioctl_admxrc3.h>
#include <admxrc3/types.h>

typedef enum _State {
  StateInitial = 0,
  StateAcquireTicket = 1,
  StateAcquireFailed = 2,
  StateDmaTransfer = 3,
  StateDmaCompleted = 4
} State;

/* Forward declarations */
static void dmaCallback(CoreTicket* pTicket, void* pCallbackContext, CoreDmaStatus dmaStatus);
static void dmaCancelCallback(DfClientRequest* pRequest, DfIoStatus reason);
static void dmaTicketAcquiredCallback(void* pDevice, CoreTicket* pTicket, CoreAcquireStatus result);

/*
** -----------------------------------------------------------------------------
** Utility code
** -----------------------------------------------------------------------------
*/

static const char*
getStateDescription(
  State state)
{
  switch (state) {
  case StateInitial:
    return "Initial";

  case StateAcquireTicket:
    return "AcquireTicket";

  case StateAcquireFailed:
    return "AcquireFailed";

  case StateDmaTransfer:
    return "DmaTransfer";

  case StateDmaCompleted:
    return "DmaCompleted";

  default:
    return "INVALID";
  }
}

static const char*
getDmaStatusDescription(
  CoreDmaStatus status)
{
  switch (status) {
  case CoreDmaSuccess:
    return "Success";

  case CoreDmaInvalidChannel:
    return "InvalidChannel";

  case CoreDmaInvalidRegion:
    return "InvalidRegion";

  case CoreDmaHardwareError:
    return "HardwareError";

  case CoreDmaUnexpectedError:
    return "UnexpectedError";

  case CoreDmaInvalidTicket:
    return "InvalidTicket";

  case CoreDmaCancelled:
    return "Cancelled";

  default:
    return "INVALID";
  }
}

static DfIoStatus
mapDescriptionResult(
  DfDescriptionResult result)
{
  switch (result) {
  case DfDescriptionSuccess:
    return DfIoStatusSuccess;

  case DfDescriptionNullPointer:
    return DfIoStatusError(ADMXRC3_NULL_POINTER);

  case DfDescriptionInvalidBuffer:
    return DfIoStatusError(ADMXRC3_INVALID_BUFFER);

  case DfDescriptionNoMemory:
    return DfIoStatusError(ADMXRC3_NO_MEMORY);

  case DfDescriptionUnexpectedError:
  default:
    return DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR);
  }
}

static DfIoStatus
mapDmaStatus(
  CoreDmaStatus dmaStatus)
{
  switch (dmaStatus) {
  case CoreDmaSuccess:
    return DfIoStatusSuccess;

  case CoreDmaHardwareError:
    return DfIoStatusError(ADMXRC3_HARDWARE_ERROR);

  case CoreDmaInvalidChannel:
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);

  case CoreDmaInvalidRegion:
    return DfIoStatusError(ADMXRC3_INVALID_REGION);

  case CoreDmaCancelled:
    return DfIoStatusError(ADMXRC3_CANCELLED);

  case CoreDmaUnexpectedError:
  case CoreDmaInvalidTicket:
  default:
    return DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR);
  }
}

static DfIoStatus
mapAcquireStatus(
  CoreAcquireStatus acqStatus)
{
  switch (acqStatus) {
  case CoreAcquireSuccess:
    return DfIoStatusSuccess;

  case CoreAcquireMustWait:
    return DfIoStatusError(ADMXRC3_MUST_WAIT);

  case CoreAcquireCancelled:
    return DfIoStatusError(ADMXRC3_CANCELLED);

  case CoreAcquireRemoved:
  case CoreAcquireFailed:
    return DfIoStatusError(ADMXRC3_HARDWARE_ERROR);

  case CoreAcquireInvalidIndex:
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);

  case CoreAcquireInvalid:
  case CoreAcquireUnexpectedError:
  case CoreAcquireWaiting:
  default:
    return DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR);
  }
}

void
completeRequest(
  Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  DmaTicket* pTicket,
  DfIoStatus status)
{
  DfClientRequest* pRequest = pTicket->pRequest;
  DfBufferDescription* pDescription;
  boolean_t bDmaLocked = pTicket->bDmaLocked;

  dfRequestComplete(pRequest, status);
  if (bDmaLocked) {
    dereferenceBufferDescription(pDevCtx, pClCtx, pTicket->variant.pBuffer);
  } else {
    pDescription = pTicket->variant.pDescription;
    dfBufferDescriptionUnlock(pDescription);
    dfBufferDescriptionDelete(pDescription);
  }
  dfPoolFree(&pDevCtx->dma.ticketPool, pTicket);
}

/*
** -----------------------------------------------------------------------------
** Callbacks
** -----------------------------------------------------------------------------
*/

static void
dmaCancelCallback(
  DfClientRequest* pRequest,
  DfIoStatus reason)
{
  DmaTicket* pTicket = (DmaTicket*)dfRequestGetContext(pRequest);
  Admxrc3DeviceContext* pDevCtx = pTicket->pDevCtx;
  Admxrc3ClientContext* pClCtx = pTicket->pClCtx;
  unsigned int dmaChannel = pTicket->dmaChannel;
  DfSpinLockFlags f;

  f = dfSpinLockGet(&pTicket->lock);
  dfDebugPrint(3, ("dmaCancelCallback: pRequest=0x%p state=%s\n", pRequest, getStateDescription(pTicket->state)));
  switch (pTicket->state) {
  case StateAcquireTicket:
    dfDebugPrint(3, ("dmaCancelCallback: cancelling acquisition\n"));
    pTicket->cancel.bCancel = TRUE;
    dfSpinLockPut(&pTicket->lock, f);
    pDevCtx->coreInterface.pAcquireCancel(pDevCtx->pCoreContext, &pTicket->ticket);
    break;

  case StateAcquireFailed:
    /* Acquisition failed, so we can complete the request here */
    dfDebugPrint(3, ("dmaCancelCallback: acquisition failed\n"));
    dfSpinLockPut(&pTicket->lock, f);
    completeRequest(pDevCtx, pClCtx, pTicket, DfIoStatusError(ADMXRC3_CANCELLED));
    break;

  case StateDmaTransfer:
    /* Cancel the DMA transfer - this will cause (abnormal) completion of the DMA sooner than would otherwise occur */
    dfDebugPrint(3, ("dmaCancelCallback: cancelling DMA transfer\n"));
    pTicket->cancel.bCancel = TRUE;
    dfSpinLockPut(&pTicket->lock, f);
    pDevCtx->coreInterface.pDmaCancel(pDevCtx->pCoreContext, &pTicket->ticket, dmaChannel);
    break;

  case StateDmaCompleted:
    /* The DMA transfer has already completed, so we can complete the request here */
    dfDebugPrint(3, ("dmaCancelCallback: DMA transfer completed\n"));
    dfSpinLockPut(&pTicket->lock, f);
    pDevCtx->coreInterface.pRelease(pDevCtx->pCoreContext, &pTicket->ticket);
    completeRequest(pDevCtx, pClCtx, pTicket, DfIoStatusError(ADMXRC3_CANCELLED));
    break;

  case StateInitial:
  default:
    /* Should never get here! */
    dfDebugPrint(0, ("dmaCancelCallback: unexpected state %s, doing nothing\n", getStateDescription(pTicket->state)));
    dfSpinLockPut(&pTicket->lock, f);
    dfAssert(FALSE);
    break;
  }
}

static void
dmaTicketAcquiredCallback(
  void* pDevice,
  CoreTicket* pCoreTicket,
  CoreAcquireStatus acqStatus)
{
  DmaTicket* pTicket = DF_CONTAINER_OF(pCoreTicket, DmaTicket, ticket);
  Admxrc3DeviceContext* pDevCtx = pTicket->pDevCtx;
  Admxrc3ClientContext* pClCtx = pTicket->pClCtx;
  DfClientRequest* pRequest = pTicket->pRequest;
  unsigned int dmaChannel;
  DfSpinLockFlags f;

  dmaChannel = pTicket->dmaChannel;
  f = dfSpinLockGet(&pTicket->lock);
  switch (acqStatus) {
  case CoreAcquireSuccess:
    if (pTicket->cancel.bCancel) {
      /* Cancel function has already run, so we should complete the request here */
      dfDebugPrint(1, ("dmaTicketAcquiredCallback: cancel function has already executed\n"));
      dfSpinLockPut(&pTicket->lock, f);
      pDevCtx->coreInterface.pRelease(pDevCtx->pCoreContext, &pTicket->ticket);
      completeRequest(pDevCtx, pClCtx, pTicket, DfIoStatusError(ADMXRC3_CANCELLED));
    } else {
      /* Acquired ticket, proceed with the DMA transfer */
      dfDebugPrint(6, ("dmaTicketAcquiredCallback: proceeding with DMA transfer\n"));
      pDevCtx->coreInterface.pDmaInitialize(pDevCtx->pCoreContext, &pTicket->ticket, dmaChannel, &pTicket->dmaParameters, dmaCallback, NULL);
      pTicket->state = StateDmaTransfer;
      dfSpinLockPut(&pTicket->lock, f);
      pDevCtx->coreInterface.pDmaStart(pDevCtx->pCoreContext, &pTicket->ticket, dmaChannel);
    }
    break;

  case CoreAcquireCancelled:
    /* We were cancelled; means that cancel function has already completed */
    dfDebugPrint(1, ("dmaTicketAcquiredCallback: we were cancelled\n"));
    pTicket->state = StateAcquireFailed;
    dfSpinLockPut(&pTicket->lock, f);
    completeRequest(pDevCtx, pClCtx, pTicket, DfIoStatusError(ADMXRC3_CANCELLED));
    break;

  default:
    if (dfRequestClearCallback(pRequest)) {
      dfDebugPrint(6, ("dmaTicketAcquiredCallback: acquisition error, cancel function removed\n"));
      pTicket->state = StateAcquireFailed;
      dfSpinLockPut(&pTicket->lock, f);
      completeRequest(pDevCtx, pClCtx, pTicket, DfIoStatusError(ADMXRC3_CANCELLED));
    } else {
      /* Cancel function is running - do nothing here */
      dfDebugPrint(1, ("dmaTicketAcquiredCallback: acquisition error, cancel function running\n"));
      pTicket->state = StateAcquireFailed;
      dfSpinLockPut(&pTicket->lock, f);
    }
  }
}

static void
dmaCallback(
  CoreTicket* pCoreTicket,
  void* pCallbackContext, /* ignored */
  CoreDmaStatus dmaStatus)
{
  DmaTicket* pTicket = DF_CONTAINER_OF(pCoreTicket, DmaTicket, ticket);
	Admxrc3DeviceContext* pDevCtx = pTicket->pDevCtx;
  Admxrc3ClientContext* pClCtx = pTicket->pClCtx;
  DfClientRequest* pRequest = pTicket->pRequest;
  DfSpinLockFlags f;

  f = dfSpinLockGet(&pTicket->lock);
  pTicket->state = StateDmaCompleted;
  if (pTicket->cancel.bCancel) {
    /* The cancel function must have already run - we can complete the request */
    dfDebugPrint(3, ("dmaCallback: cancel function has already executed\n"));
    dfSpinLockPut(&pTicket->lock, f);
    pDevCtx->coreInterface.pRelease(pDevCtx->pCoreContext, &pTicket->ticket);
    completeRequest(pDevCtx, pClCtx, pTicket, DfIoStatusError(ADMXRC3_CANCELLED));
  } else {
    if (dfRequestClearCallback(pRequest)) {
      /* We successfully removed the cancel function - we can complete the request */
      dfDebugPrint(3, ("dmaCallback: removed cancel function, pRequest=0x%p dmaStatus=%s\n",
        pRequest, getDmaStatusDescription(dmaStatus)));
      dfSpinLockPut(&pTicket->lock, f);
      pDevCtx->coreInterface.pRelease(pDevCtx->pCoreContext, &pTicket->ticket);
      completeRequest(pDevCtx, pClCtx, pTicket, mapDmaStatus(dmaStatus));
    } else {
      /* The cancel function must be running - let it complete the request */
      dfDebugPrint(3, ("dmaCallback: cancel function is running\n"));
      dfSpinLockPut(&pTicket->lock, f);
    }
  }
}

/*
** -----------------------------------------------------------------------------
** IOCTL handlers
** -----------------------------------------------------------------------------
*/

#if defined(ADMXRC3_CHECK_FLAGS)
static const uint32_t legalFlagsMask = ADMXRC3_DMA_FIXEDLOCAL | ADMXRC3_DMA_DONOTQUEUE;
#endif

DfIoStatus
ioctlReadDma(
  Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  DfClientRequest* pRequest,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS_ADMXRC3_READDMA* pIoctl = (IOCTLS_ADMXRC3_READDMA*)pBuffer;
  DmaTicket* pTicket = NULL;
  CoreAcquireStatus acqStatus;
  unsigned int dmaChannel;
  DfSpinLockFlags f;
  DfBufferDescription* pDescription;
  DfDescriptionResult descResult;
  size_t length;
  uint32_t acquireFlags, flags, localAddress;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }
  if (!pClCtx->bPrivileged) {
    return DfIoStatusError(ADMXRC3_ACCESS_DENIED);
  }

  dmaChannel = pIoctl->in.channel;
  if (dmaChannel >= pDevCtx->coreInterface.pGetNumDmaChannel(pDevCtx->pCoreContext)) {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }

  length = pIoctl->in.length;
  if (0 == length) {
    dfRequestPending(pRequest);
    dfRequestComplete(pRequest, DfIoStatusSuccess);
    return DfIoStatusDeferred;
  }
  flags = pIoctl->in.flags;
  localAddress = pIoctl->in.local;
#if defined(ADMXRC3_CHECK_FLAGS)
  if (0 != (flags & ~legalFlagsMask)) {
    return DfIoStatusError(ADMXRC3_INVALID_FLAG);
  }
#endif
  if (!(flags & ADMXRC3_DMA_FIXEDLOCAL)) {
    if (length > 0xffffffffU || localAddress + (uint32_t)length < localAddress) {
      return DfIoStatusError(ADMXRC3_INVALID_LOCAL_REGION);
    }
  }
  acquireFlags = (flags & ADMXRC3_DMA_DONOTQUEUE) ? CORE_ACQUIRE_NOWAIT : 0;
  descResult = dfBufferDescriptionCreateUser(pIoctl->in.pBuffer, length, &pDescription);
  if (DfDescriptionSuccess != descResult) {
    return mapDescriptionResult(descResult);
  }
  if (!dfBufferDescriptionLock(pDescription)) {
    dfBufferDescriptionDelete(pDescription);
    return DfIoStatusError(ADMXRC3_INVALID_BUFFER);
  }

  pTicket = dfPoolAlloc(&pDevCtx->dma.ticketPool, DmaTicket);
  if (NULL == pTicket) {
    dfBufferDescriptionUnlock(pDescription);
    dfBufferDescriptionDelete(pDescription);
    return DfIoStatusError(ADMXRC3_NO_MEMORY);
  }

  dfSpinLockInit(&pTicket->lock);
  pTicket->pRequest = pRequest;
  pTicket->pDevCtx = pDevCtx;
  pTicket->pClCtx = pClCtx;
  pTicket->state = StateInitial;
  pTicket->dmaChannel = dmaChannel;
  pTicket->bDmaLocked = FALSE;
  pTicket->variant.pDescription = pDescription;
  pTicket->dmaParameters.pBuffer = pDescription;
  pTicket->dmaParameters.bWriteToLocal = FALSE;
  pTicket->dmaParameters.bFixedLocalAddress = (flags & ADMXRC3_DMA_FIXEDLOCAL) ? TRUE : FALSE;
  pTicket->dmaParameters.region.offset = 0;
  pTicket->dmaParameters.region.length = length;
  pTicket->dmaParameters.region.localAddress = localAddress;
  pTicket->cancel.bCancel = FALSE;
  pTicket->ticket.count = 1;
  pTicket->ticket.resources[0].code = CORE_RESOURCE_DMAENGINE;
  pTicket->ticket.resources[0].index = (uint8_t)dmaChannel;
  dfRequestSetContext(pRequest, pTicket);

  acqStatus = pDevCtx->coreInterface.pAcquireInitialize(pDevCtx->pCoreContext, &pTicket->ticket);
  if (acqStatus != CoreAcquireSuccess) {
    dfPoolFree(&pDevCtx->dma.ticketPool, pTicket);
    dfBufferDescriptionUnlock(pDescription);
    dfBufferDescriptionDelete(pDescription);
    return mapAcquireStatus(acqStatus);
  }

  f = dfSpinLockGet(&pTicket->lock);
  if (!dfRequestSetCallback(pRequest, dmaCancelCallback)) {
    dfSpinLockPut(&pTicket->lock, f);
    dfPoolFree(&pDevCtx->dma.ticketPool, pTicket);
    dfBufferDescriptionUnlock(pDescription);
    dfBufferDescriptionDelete(pDescription);
    return DfIoStatusError(ADMXRC3_CANCELLED);
  }
  pTicket->state = StateAcquireTicket;
  dfSpinLockPut(&pTicket->lock, f);

  dfRequestPending(pRequest);
  pDevCtx->coreInterface.pAcquireAsync(pDevCtx->pCoreContext, &pTicket->ticket, acquireFlags, dmaTicketAcquiredCallback);

  return DfIoStatusDeferred;
}

DfIoStatus
ioctlReadDmaLocked(
  Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  DfClientRequest* pRequest,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS_ADMXRC3_READDMALOCKED* pIoctl = (IOCTLS_ADMXRC3_READDMALOCKED*)pBuffer;
  DmaTicket* pTicket = NULL;
  CoreAcquireStatus acqStatus;
  unsigned int dmaChannel;
  DfSpinLockFlags f;
  BufferDescription* pUserBuffer;
  DfBufferDescription* pDescription;
  size_t offset, length;
  uint32_t acquireFlags, flags, handle, localAddress;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }
  if (!pClCtx->bPrivileged) {
    return DfIoStatusError(ADMXRC3_ACCESS_DENIED);
  }

  dmaChannel = pIoctl->in.channel;
  if (dmaChannel >= pDevCtx->coreInterface.pGetNumDmaChannel(pDevCtx->pCoreContext)) {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }

  handle = pIoctl->in.hBuffer;
  offset = pIoctl->in.offset;
  length = pIoctl->in.length;
  if (0 == length) {
    dfRequestPending(pRequest);
    dfRequestComplete(pRequest, DfIoStatusSuccess);
    return DfIoStatusDeferred;
  }
  if (offset + length < offset) {
    return DfIoStatusError(ADMXRC3_INVALID_REGION);
  }
  flags = pIoctl->in.flags;
  localAddress = pIoctl->in.local;
#if defined(ADMXRC3_CHECK_FLAGS)
  if (0 != (flags & ~legalFlagsMask)) {
    return DfIoStatusError(ADMXRC3_INVALID_FLAG);
  }
#endif
  if (!(flags & ADMXRC3_DMA_FIXEDLOCAL)) {
    if (length > 0xffffffffU || localAddress + (uint32_t)length < localAddress) {
      return DfIoStatusError(ADMXRC3_INVALID_LOCAL_REGION);
    }
  }
  acquireFlags = (flags & ADMXRC3_DMA_DONOTQUEUE) ? CORE_ACQUIRE_NOWAIT : 0;
  pUserBuffer = referenceBufferHandle(pDevCtx, pClCtx, handle - 1);
  if (NULL == pUserBuffer) {
    return DfIoStatusError(ADMXRC3_INVALID_BUFFER_HANDLE);
  }
  pDescription = pUserBuffer->pDfBufferDesc;
  if (offset + length > pDescription->length) {
    dereferenceBufferDescription(pDevCtx, pClCtx, pUserBuffer);
    return DfIoStatusError(ADMXRC3_INVALID_REGION);
  }

  pTicket = dfPoolAlloc(&pDevCtx->dma.ticketPool, DmaTicket);
  if (NULL == pTicket) {
    dereferenceBufferDescription(pDevCtx, pClCtx, pUserBuffer);
    return DfIoStatusError(ADMXRC3_NO_MEMORY);
  }

  dfSpinLockInit(&pTicket->lock);
  pTicket->pRequest = pRequest;
  pTicket->pDevCtx = pDevCtx;
  pTicket->pClCtx = pClCtx;
  pTicket->state = StateInitial;
  pTicket->dmaChannel = dmaChannel;
  pTicket->bDmaLocked = TRUE;
  pTicket->variant.pBuffer = pUserBuffer;
  pTicket->dmaParameters.pBuffer = pDescription;
  pTicket->dmaParameters.bFixedLocalAddress = (flags & ADMXRC3_DMA_FIXEDLOCAL) ? TRUE : FALSE;
  pTicket->dmaParameters.bWriteToLocal = FALSE;
  pTicket->dmaParameters.region.offset = offset;
  pTicket->dmaParameters.region.length = length;
  pTicket->dmaParameters.region.localAddress = localAddress;
  pTicket->cancel.bCancel = FALSE;
  pTicket->ticket.count = 1;
  pTicket->ticket.resources[0].code = CORE_RESOURCE_DMAENGINE;
  pTicket->ticket.resources[0].index = (uint8_t)dmaChannel;
  dfRequestSetContext(pRequest, pTicket);

  acqStatus = pDevCtx->coreInterface.pAcquireInitialize(pDevCtx->pCoreContext, &pTicket->ticket);
  if (acqStatus != CoreAcquireSuccess) {
    dfPoolFree(&pDevCtx->dma.ticketPool, pTicket);
    dereferenceBufferDescription(pDevCtx, pClCtx, pUserBuffer);
    return mapAcquireStatus(acqStatus);
  }

  f = dfSpinLockGet(&pTicket->lock);
  if (!dfRequestSetCallback(pRequest, dmaCancelCallback)) {
    dfSpinLockPut(&pTicket->lock, f);
    dfPoolFree(&pDevCtx->dma.ticketPool, pTicket);
    dereferenceBufferDescription(pDevCtx, pClCtx, pUserBuffer);
    return DfIoStatusError(ADMXRC3_CANCELLED);
  }
  pTicket->state = StateAcquireTicket;
  dfSpinLockPut(&pTicket->lock, f);

  dfRequestPending(pRequest);
  pDevCtx->coreInterface.pAcquireAsync(pDevCtx->pCoreContext, &pTicket->ticket, acquireFlags, dmaTicketAcquiredCallback);

  return DfIoStatusDeferred;
}

#if DF_NEED_THUNK
DfIoStatus
ioctlReadDmaLockedThunk(
  Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  DfClientRequest* pRequest,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_READDMALOCKED* pIoctl32 = (IOCTLS32_ADMXRC3_READDMALOCKED*)pBuffer;
  IOCTLS_ADMXRC3_READDMALOCKED ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.channel = pIoctl32->in.channel;
  ioctl.in.flags = pIoctl32->in.flags;
  ioctl.in.hBuffer = pIoctl32->in.hBuffer;
  ioctl.in.offset = (size_t)pIoctl32->in.offset;
  ioctl.in.length = (size_t)pIoctl32->in.length;
  ioctl.in.local = pIoctl32->in.local;
  return ioctlReadDmaLocked(pDevCtx, pClCtx, pRequest, &ioctl, sizeof(ioctl.in));
}
#endif

#if DF_NEED_THUNK
DfIoStatus
ioctlReadDmaThunk(
  Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  DfClientRequest* pRequest,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_READDMA* pIoctl32 = (IOCTLS32_ADMXRC3_READDMA*)pBuffer;
  IOCTLS_ADMXRC3_READDMA ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.channel = pIoctl32->in.channel;
  ioctl.in.flags = pIoctl32->in.flags;
  ioctl.in.pBuffer = dfThunkPtr(pIoctl32->in.pBuffer);
  ioctl.in.length = (size_t)pIoctl32->in.length;
  ioctl.in.local = pIoctl32->in.local;
  return ioctlReadDma(pDevCtx, pClCtx, pRequest, &ioctl, sizeof(ioctl.in));
}
#endif

DfIoStatus
ioctlWriteDma(
  Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  DfClientRequest* pRequest,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS_ADMXRC3_WRITEDMA* pIoctl = (IOCTLS_ADMXRC3_WRITEDMA*)pBuffer;
  DmaTicket* pTicket = NULL;
  CoreAcquireStatus acqStatus;
  unsigned int dmaChannel;
  DfSpinLockFlags f;
  DfBufferDescription* pDescription;
  DfDescriptionResult descResult;
  size_t length;
  uint32_t acquireFlags, flags, localAddress;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }
  if (!pClCtx->bPrivileged) {
    return DfIoStatusError(ADMXRC3_ACCESS_DENIED);
  }

  dmaChannel = pIoctl->in.channel;
  if (dmaChannel >= pDevCtx->coreInterface.pGetNumDmaChannel(pDevCtx->pCoreContext)) {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }

  length = pIoctl->in.length;
  if (0 == length) {
    dfRequestPending(pRequest);
    dfRequestComplete(pRequest, DfIoStatusSuccess);
    return DfIoStatusDeferred;
  }
  flags = pIoctl->in.flags;
  localAddress = pIoctl->in.local;
#if defined(ADMXRC3_CHECK_FLAGS)
  if (0 != (flags & ~legalFlagsMask)) {
    return DfIoStatusError(ADMXRC3_INVALID_FLAG);
  }
#endif
  if (!(flags & ADMXRC3_DMA_FIXEDLOCAL)) {
    if (length > 0xffffffffU || localAddress + (uint32_t)length < localAddress) {
      return DfIoStatusError(ADMXRC3_INVALID_LOCAL_REGION);
    }
  }
  acquireFlags = (flags & ADMXRC3_DMA_DONOTQUEUE) ? CORE_ACQUIRE_NOWAIT : 0;
  descResult = dfBufferDescriptionCreateUser((void*)pIoctl->in.pBuffer, length, &pDescription);
  if (DfDescriptionSuccess != descResult) {
    return mapDescriptionResult(descResult);
  }
  if (!dfBufferDescriptionLock(pDescription)) {
    dfBufferDescriptionDelete(pDescription);
    return DfIoStatusError(ADMXRC3_INVALID_BUFFER);
  }

  pTicket = dfPoolAlloc(&pDevCtx->dma.ticketPool, DmaTicket);
  if (NULL == pTicket) {
    dfBufferDescriptionUnlock(pDescription);
    dfBufferDescriptionDelete(pDescription);
    return DfIoStatusError(ADMXRC3_NO_MEMORY);
  }

  dfSpinLockInit(&pTicket->lock);
  pTicket->pRequest = pRequest;
  pTicket->pDevCtx = pDevCtx;
  pTicket->pClCtx = pClCtx;
  pTicket->state = StateInitial;
  pTicket->dmaChannel = dmaChannel;
  pTicket->bDmaLocked = FALSE;
  pTicket->variant.pDescription = pDescription;
  pTicket->dmaParameters.pBuffer = pDescription;
  pTicket->dmaParameters.bFixedLocalAddress = (flags & ADMXRC3_DMA_FIXEDLOCAL) ? TRUE : FALSE;
  pTicket->dmaParameters.bWriteToLocal = TRUE;
  pTicket->dmaParameters.region.offset = 0;
  pTicket->dmaParameters.region.length = length;
  pTicket->dmaParameters.region.localAddress = localAddress;
  pTicket->cancel.bCancel = FALSE;
  pTicket->ticket.count = 1;
  pTicket->ticket.resources[0].code = CORE_RESOURCE_DMAENGINE;
  pTicket->ticket.resources[0].index = (uint8_t)dmaChannel;
  dfRequestSetContext(pRequest, pTicket);

  acqStatus = pDevCtx->coreInterface.pAcquireInitialize(pDevCtx->pCoreContext, &pTicket->ticket);
  if (acqStatus != CoreAcquireSuccess) {
    dfPoolFree(&pDevCtx->dma.ticketPool, pTicket);
    dfBufferDescriptionUnlock(pDescription);
    dfBufferDescriptionDelete(pDescription);
    return mapAcquireStatus(acqStatus);
  }

  f = dfSpinLockGet(&pTicket->lock);
  if (!dfRequestSetCallback(pRequest, dmaCancelCallback)) {
    dfSpinLockPut(&pTicket->lock, f);
    dfPoolFree(&pDevCtx->dma.ticketPool, pTicket);
    dfBufferDescriptionUnlock(pDescription);
    dfBufferDescriptionDelete(pDescription);
    return DfIoStatusError(ADMXRC3_CANCELLED);
  }
  pTicket->state = StateAcquireTicket;
  dfSpinLockPut(&pTicket->lock, f);

  dfRequestPending(pRequest);
  pDevCtx->coreInterface.pAcquireAsync(pDevCtx->pCoreContext, &pTicket->ticket, acquireFlags, dmaTicketAcquiredCallback);

  return DfIoStatusDeferred;
}

DfIoStatus
ioctlWriteDmaLocked(
  Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  DfClientRequest* pRequest,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS_ADMXRC3_WRITEDMALOCKED* pIoctl = (IOCTLS_ADMXRC3_WRITEDMALOCKED*)pBuffer;
  DmaTicket* pTicket = NULL;
  CoreAcquireStatus acqStatus;
  unsigned int dmaChannel;
  DfSpinLockFlags f;
  BufferDescription* pUserBuffer;
  DfBufferDescription* pDescription;
  size_t offset, length;
  uint32_t acquireFlags, flags, handle, localAddress;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }
  if (!pClCtx->bPrivileged) {
    return DfIoStatusError(ADMXRC3_ACCESS_DENIED);
  }

  dmaChannel = pIoctl->in.channel;
  if (dmaChannel >= pDevCtx->coreInterface.pGetNumDmaChannel(pDevCtx->pCoreContext)) {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }

  handle = pIoctl->in.hBuffer;
  offset = pIoctl->in.offset;
  length = pIoctl->in.length;
  if (0 == length) {
    dfRequestPending(pRequest);
    dfRequestComplete(pRequest, DfIoStatusSuccess);
    return DfIoStatusDeferred;
  }
  if (offset + length < offset) {
    return DfIoStatusError(ADMXRC3_INVALID_REGION);
  }
  flags = pIoctl->in.flags;
  localAddress = pIoctl->in.local;
#if defined(ADMXRC3_CHECK_FLAGS)
  if (0 != (flags & ~legalFlagsMask)) {
    return DfIoStatusError(ADMXRC3_INVALID_FLAG);
  }
#endif
  if (!(flags & ADMXRC3_DMA_FIXEDLOCAL)) {
    if (length > 0xffffffffU || localAddress + (uint32_t)length < localAddress) {
      return DfIoStatusError(ADMXRC3_INVALID_LOCAL_REGION);
    }
  }
  acquireFlags = (flags & ADMXRC3_DMA_DONOTQUEUE) ? CORE_ACQUIRE_NOWAIT : 0;
  pUserBuffer = referenceBufferHandle(pDevCtx, pClCtx, handle - 1);
  if (NULL == pUserBuffer) {
    return DfIoStatusError(ADMXRC3_INVALID_BUFFER_HANDLE);
  }
  pDescription = pUserBuffer->pDfBufferDesc;
  if (offset + length > pDescription->length) {
    dereferenceBufferDescription(pDevCtx, pClCtx, pUserBuffer);
    return DfIoStatusError(ADMXRC3_INVALID_REGION);
  }

  pTicket = dfPoolAlloc(&pDevCtx->dma.ticketPool, DmaTicket);
  if (NULL == pTicket) {
    dereferenceBufferDescription(pDevCtx, pClCtx, pUserBuffer);
    return DfIoStatusError(ADMXRC3_NO_MEMORY);
  }

  dfSpinLockInit(&pTicket->lock);
  pTicket->pRequest = pRequest;
  pTicket->pDevCtx = pDevCtx;
  pTicket->pClCtx = pClCtx;
  pTicket->state = StateInitial;
  pTicket->dmaChannel = dmaChannel;
  pTicket->bDmaLocked = TRUE;
  pTicket->variant.pBuffer = pUserBuffer;
  pTicket->dmaParameters.pBuffer = pDescription;
  pTicket->dmaParameters.bFixedLocalAddress = (flags & ADMXRC3_DMA_FIXEDLOCAL) ? TRUE : FALSE;
  pTicket->dmaParameters.bWriteToLocal = TRUE;
  pTicket->dmaParameters.region.offset = offset;
  pTicket->dmaParameters.region.length = length;
  pTicket->dmaParameters.region.localAddress = localAddress;
  pTicket->cancel.bCancel = FALSE;
  pTicket->ticket.count = 1;
  pTicket->ticket.resources[0].code = CORE_RESOURCE_DMAENGINE;
  pTicket->ticket.resources[0].index = (uint8_t)dmaChannel;
  dfRequestSetContext(pRequest, pTicket);

  acqStatus = pDevCtx->coreInterface.pAcquireInitialize(pDevCtx->pCoreContext, &pTicket->ticket);
  if (acqStatus != CoreAcquireSuccess) {
    dfPoolFree(&pDevCtx->dma.ticketPool, pTicket);
    dereferenceBufferDescription(pDevCtx, pClCtx, pUserBuffer);
    return mapAcquireStatus(acqStatus);
  }

  f = dfSpinLockGet(&pTicket->lock);
  if (!dfRequestSetCallback(pRequest, dmaCancelCallback)) {
    dfSpinLockPut(&pTicket->lock, f);
    dfPoolFree(&pDevCtx->dma.ticketPool, pTicket);
    dereferenceBufferDescription(pDevCtx, pClCtx, pUserBuffer);
    return DfIoStatusError(ADMXRC3_CANCELLED);
  }
  pTicket->state = StateAcquireTicket;
  dfSpinLockPut(&pTicket->lock, f);

  dfRequestPending(pRequest);
  pDevCtx->coreInterface.pAcquireAsync(pDevCtx->pCoreContext, &pTicket->ticket, acquireFlags, dmaTicketAcquiredCallback);

  return DfIoStatusDeferred;
}

#if DF_NEED_THUNK
DfIoStatus
ioctlWriteDmaLockedThunk(
  Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  DfClientRequest* pRequest,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_WRITEDMALOCKED* pIoctl32 = (IOCTLS32_ADMXRC3_WRITEDMALOCKED*)pBuffer;
  IOCTLS_ADMXRC3_WRITEDMALOCKED ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.channel = pIoctl32->in.channel;
  ioctl.in.flags = pIoctl32->in.flags;
  ioctl.in.hBuffer = pIoctl32->in.hBuffer;
  ioctl.in.offset = (size_t)pIoctl32->in.offset;
  ioctl.in.length = (size_t)pIoctl32->in.length;
  ioctl.in.local = pIoctl32->in.local;
  return ioctlWriteDmaLocked(pDevCtx, pClCtx, pRequest, &ioctl, sizeof(ioctl.in));
}
#endif

#if DF_NEED_THUNK
DfIoStatus
ioctlWriteDmaThunk(
  Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx,
  DfClientRequest* pRequest,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_WRITEDMA* pIoctl32 = (IOCTLS32_ADMXRC3_WRITEDMA*)pBuffer;
  IOCTLS_ADMXRC3_WRITEDMA ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.channel = pIoctl32->in.channel;
  ioctl.in.flags = pIoctl32->in.flags;
  ioctl.in.pBuffer = dfThunkPtr(pIoctl32->in.pBuffer);
  ioctl.in.length = (size_t)pIoctl32->in.length;
  ioctl.in.local = pIoctl32->in.local;
  return ioctlWriteDma(pDevCtx, pClCtx, pRequest, &ioctl, sizeof(ioctl.in));
}
#endif

/*
** -----------------------------------------------------------------------------
** Cleanup code; called when a device handle is being closed
** -----------------------------------------------------------------------------
*/

void
cleanupDma(
  Admxrc3DeviceContext* pDevCtx,
  Admxrc3ClientContext* pClCtx)
{
  dfDebugPrint(1, ("cleanupDma: entered, pDevCtx=0x%p pClCtx=0x%p\n", pDevCtx, pClCtx)); 
}
