/*
** File: fpga.c  
** Project: ADMXRC3 module driver
** Purpose: OS-independent IOCTL handlers for configuring target FPGA(s).
**
** (C) Copyright Alpha Data 2009
*/

#include <df.h>
#include "device.h"
#include "fpga.h"
#include <admxrc3/types.h>

static void
reverseBytes(
  void* pBuf,
  size_t n)
{
  const size_t mask32 = 0x3U;
  uint8_t* pBuf8;
  uint8_t a8, b8;
  uint32_t a32, b32;
  uint32_t* pBuf32;
  size_t n8, n32;

  /* Ensure pointer aligned to 4-byte boundary by doing first few bytes */
  pBuf8 = (uint8_t*)pBuf;
  n8 = (size_t)(uintptr_t)pBuf & mask32;
  n -= n8;
  while (n8) {
    a8 = *pBuf8;
    /* Reverse bits in byte */
    b8 = ((a8 & 0x01U) << 7) | ((a8 & 0x02U) << 5) | ((a8 & 0x04U) << 3) | ((a8 & 0x08U) << 1) |
      ((a8 & 0x10U) >> 1) | ((a8 & 0x20U) >> 3) | ((a8 & 0x40U) >> 5) | ((a8 & 0x80U) >> 7);
    *pBuf8++ = b8;
  }

  /* Do bulk of bit-reversal as 32-bit words */
  pBuf32 = (uint32_t*)pBuf8;
  n32 = n & ~mask32;
  n -= n32;
  while (n32) {
    a32 = *pBuf32;
    /* Reverse bits in each byte of 32-bit word */
    b32 = ((a32 & 0x01010101U) << 7) | ((a32 & 0x02020202U) << 5) | ((a32 & 0x04040404U) << 3) | ((a32 & 0x08080808U) << 1) |
      ((a32 & 0x10101010U) >> 1) | ((a32 & 0x20202020U) >> 3) | ((a32 & 0x40404040U) >> 5) | ((a32 & 0x80808080U) >> 7);
    *pBuf32++ = b32;
    n32 -= 4;
  }

  /* Do last few bytes */
  pBuf8 = (uint8_t*)pBuf32;
  while (n--) {
    a8 = *pBuf8;
    /* Reverse bits in byte */
    b8 = ((a8 & 0x01U) << 7) | ((a8 & 0x02U) << 5) | ((a8 & 0x04U) << 3) | ((a8 & 0x08U) << 1) |
      ((a8 & 0x10U) >> 1) | ((a8 & 0x20U) >> 3) | ((a8 & 0x40U) >> 5) | ((a8 & 0x80U) >> 7);
    *pBuf8++ = b8;
  }
}

static DfIoStatus
mapAcquireStatus(
  CoreAcquireStatus acqStatus)
{
  switch (acqStatus) {
  case CoreAcquireSuccess:
    return DfIoStatusSuccess;

  case CoreAcquireMustWait:
    return DfIoStatusError(ADMXRC3_MUST_WAIT);

  case CoreAcquireCancelled:
    return DfIoStatusError(ADMXRC3_CANCELLED);

  case CoreAcquireRemoved:
  case CoreAcquireFailed:
    return DfIoStatusError(ADMXRC3_HARDWARE_ERROR);

  case CoreAcquireInvalidIndex:
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);

  case CoreAcquireInvalid:
  case CoreAcquireUnexpectedError:
  case CoreAcquireWaiting:
  default:
    return DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR);
  }
}

#define MAX_CHUNK (128)

DfIoStatus
ioctlConfigure(
	Admxrc3DeviceContext* pDevCtx,
	Admxrc3ClientContext* pClCtx,
  void* pBuffer,
  unsigned int inSize)
{
#if defined(ADMXRC3_CHECK_FLAGS)
  const uint32_t legalFlagsMask = ADMXRC3_CONFIGURE_PARTIAL | ADMXRC3_CONFIGURE_NOCHECK | ADMXRC3_CONFIGURE_SHARE;
#endif
  CoreTicket ticket;
  CoreTicket* pTicket = &ticket;
  uint8_t tmp[MAX_CHUNK];
  IOCTLS_ADMXRC3_CONFIGURE* pIoctl = (IOCTLS_ADMXRC3_CONFIGURE*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  DfSpinLockFlags f;
  CoreAcquireStatus acqStatus;
  unsigned int targetIndex;
  uint32_t flags;
  size_t length, remaining, chunk;
  uint8_t* pUserBuffer;
  boolean_t bInterrupt, bValue, bAcquired = FALSE;
  DfTime start, expires;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }
  if (!pClCtx->bPrivileged) {
    return DfIoStatusError(ADMXRC3_ACCESS_DENIED);
  }

  targetIndex = pIoctl->in.targetIndex;
  flags = pIoctl->in.flags;
  pUserBuffer = (uint8_t*)pIoctl->in.pBuffer;
  length = pIoctl->in.length;
  
  dfDebugPrint(6, ("ioctlConfigure: targetIndex=%lu flags=0x%lx pUserBuffer=0x%p length=%llu(0x%08lx_%08lx)\n",
    (unsigned long)targetIndex, (unsigned long)flags, pUserBuffer, (unsigned long long)length, dfSplitUint64((uint64_t)length)));

  if (targetIndex >= pDevCtx->info.numTargetFpga) {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }
#if defined(ADMXRC3_CHECK_FLAGS)
  if (flags & ~legalFlagsMask) {
    return DfIoStatusError(ADMXRC3_INVALID_FLAG);
  }
#endif

  f = dfSpinLockGet(&pDevCtx->fpga[targetIndex].owner.lock);
  if (pDevCtx->fpga[targetIndex].owner.pClCtx == NULL) {
    /* Target FPGA is currently not owned by any client */
    if (!(flags & ADMXRC3_CONFIGURE_SHARE)) {
      /* Set owner of target FPGA to this client, and proceed. */
      pDevCtx->fpga[targetIndex].owner.pClCtx = pClCtx;
    }
  } else if (pClCtx == pDevCtx->fpga[targetIndex].owner.pClCtx) {
    /* Target FPGA is currently owned by this client; OK to proceed. */
  } else {
    /* Target FPGA is currently owned by a different client; don't proceed. */
    dfSpinLockPut(&pDevCtx->fpga[targetIndex].owner.lock, f);
    return DfIoStatusError(ADMXRC3_NOT_OWNER);
  }
  dfSpinLockPut(&pDevCtx->fpga[targetIndex].owner.lock, f);

  pTicket->count = 1;
  pTicket->resources[0].code = CORE_RESOURCE_TARGETFPGA;
  pTicket->resources[0].index = (uint8_t)targetIndex;
  acqStatus = pDevCtx->coreInterface.pAcquireInitialize(pDevCtx->pCoreContext, pTicket);
  if (CoreAcquireSuccess != acqStatus) {
    status = mapAcquireStatus(acqStatus);
    goto done;
  }
  acqStatus = pDevCtx->coreInterface.pAcquireSync(pDevCtx->pCoreContext, pTicket, 0);
  if (CoreAcquireSuccess != acqStatus) {
    status = mapAcquireStatus(acqStatus);
    goto done;
  }
  bAcquired = TRUE;

  if (!(flags & ADMXRC3_CONFIGURE_PARTIAL)) {
    /* The first phase of full configuration must complete within 1s */
    start = dfTimeGet();
    expires = start + dfMicrosecondsToTime(1000000);

    /* Assert PROG */
    bValue = TRUE;
    pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpSetProg, &bValue);

    /* Wait for DONE deasserted */
    while (1) {
      pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpGetDone, &bValue);
      if (!bValue) {
        break;
      }

      bInterrupt = dfDelayThreadFor(dfMicrosecondsToTime(10000));
      if (bInterrupt) {
        /* Deassert PROG */
        bValue = FALSE;
        pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpSetProg, &bValue);

        /* Abort early */
        status = DfIoStatusError(ADMXRC3_CANCELLED);
        goto done;
      }
      if (dfTimeGet() >= expires) {
        /* Timeout waiting for DONE deasserted */
        bValue = FALSE;
        pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpSetProg, &bValue);
        status = DfIoStatusError(ADMXRC3_HARDWARE_ERROR);
        goto done;
      }
    }

    /* Wait for INIT# asserted */
    while (1) {
      pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpGetInit, &bValue);
      if (bValue) {
        break;
      }

      bInterrupt = dfDelayThreadFor(dfMicrosecondsToTime(10000));
      if (bInterrupt) {
        /* Deassert PROG */
        bValue = FALSE;
        pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpSetProg, &bValue);

        /* Abort early */
        status = DfIoStatusError(ADMXRC3_CANCELLED);
        goto done;
      }
      if (dfTimeGet() >= expires) {
        /* Timeout waiting for INIT# asserted */
        bValue = FALSE;
        pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpSetProg, &bValue);
        status = DfIoStatusError(ADMXRC3_HARDWARE_ERROR);
        goto done;
      }
    }

    /* Deassert PROG */
    bValue = FALSE;
    pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpSetProg, &bValue);

    /* Wait for INIT# deasserted */
    while (1) {
      pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpGetInit, &bValue);
      if (!bValue) {
        break;
      }

      bInterrupt = dfDelayThreadFor(dfMicrosecondsToTime(10000));
      if (bInterrupt) {
        /* Abort early */
        status = DfIoStatusError(ADMXRC3_CANCELLED);
        goto done;
      }
      if (dfTimeGet() >= expires) {
        /* Timeout waiting for INIT# deasserted */
        status = DfIoStatusError(ADMXRC3_HARDWARE_ERROR);
        goto done;
      }
    }
  }

  DF_USER_SPACE_ACCESS_BEGIN

  dfUserBufferValidate(pUserBuffer, length, FALSE);
  remaining = length;
  while (remaining) {
    chunk = MAX_CHUNK;
    if (chunk > remaining) {
      chunk = remaining;
    }
    dfCopyUK(tmp, pUserBuffer, chunk);
    if (pDevCtx->bReverseSelectMapData) {
  		reverseBytes(tmp, chunk);
    }
    if (remaining == length) {
      dfDebugPrint(8, ("ioctlConfigure: First %lu bytes of SelectMap data:\n", (unsigned long)chunk));
      dfDebugDumpMemory(8, 0, 16, 2, TRUE, tmp, chunk);
    }
    pDevCtx->coreInterface.pFpgaSelectMap(pDevCtx->pCoreContext, pTicket, targetIndex, TRUE, chunk, tmp);
    remaining -= chunk;
    pUserBuffer += chunk;
  }

  DF_USER_SPACE_ACCESS_END({
    status = DfIoStatusError(ADMXRC3_INVALID_BUFFER);
  })

  if (DfIoStatusIsError(status)) {
    goto done;
  }

  if (!(flags & ADMXRC3_CONFIGURE_PARTIAL)) {
    /* The second phase of full configuration must complete within 1s */
    start = dfTimeGet();
    expires = start + dfMicrosecondsToTime(1000000);

    /* Check if INIT# is asserted */
    pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpGetInit, &bValue);
    if (bValue) {
      status = DfIoStatusError(ADMXRC3_HARDWARE_ERROR);
      goto done;
    }

    /* Wait for DONE asserted */
    while (1) {
      pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpGetDone, &bValue);
      if (bValue) {
        break;
      }

      bInterrupt = dfDelayThreadFor(dfMicrosecondsToTime(10000));
      if (bInterrupt) {
        /* Abort early */
        status = DfIoStatusError(ADMXRC3_CANCELLED);
        goto done;
      }
      if (dfTimeGet() >= expires) {
        /* Timeout waiting for DONE asserted */
        status = DfIoStatusError(ADMXRC3_HARDWARE_ERROR);
        goto done;
      }
    }

    if (!(flags & ADMXRC3_CONFIGURE_NOCHECK)) {
      /* Wait for bridge <-> target link OK indication */
      while (1) {
        pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpGetLinkStatus, &bValue);
        if (bValue) {
          break;
        }

        bInterrupt = dfDelayThreadFor(dfMicrosecondsToTime(10000));
        if (bInterrupt) {
          /* Abort early */
          status = DfIoStatusError(ADMXRC3_CANCELLED);
          goto done;
        }
        if (dfTimeGet() >= expires) {
          /* Timeout waiting for DONE asserted */
          status = DfIoStatusError(ADMXRC3_HARDWARE_ERROR);
          goto done;
        }
      }
    }
  }

done:
  if (bAcquired) {
    pDevCtx->coreInterface.pRelease(pDevCtx->pCoreContext, pTicket);
    bAcquired = FALSE;
  }

  if (DfIoStatusIsError(status) && !(flags & ADMXRC3_CONFIGURE_SHARE)) {
    /* Failed to configure target FPGA, so mark it as not owned by any client. */
    f = dfSpinLockGet(&pDevCtx->fpga[targetIndex].owner.lock);
    pDevCtx->fpga[targetIndex].owner.pClCtx = NULL;
    dfSpinLockPut(&pDevCtx->fpga[targetIndex].owner.lock, f);
  }

  return status;
}

#if DF_NEED_THUNK
DfIoStatus
ioctlConfigureThunk(
	Admxrc3DeviceContext* pDevCtx,
	Admxrc3ClientContext* pClCtx,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_CONFIGURE* pIoctl32 = (IOCTLS32_ADMXRC3_CONFIGURE*)pBuffer;
  IOCTLS_ADMXRC3_CONFIGURE ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.targetIndex = pIoctl32->in.targetIndex;
  ioctl.in.flags = pIoctl32->in.flags;
  ioctl.in.pBuffer = dfThunkConstPtr(pIoctl32->in.pBuffer);
  ioctl.in.length = (size_t)pIoctl32->in.length;
  return ioctlConfigure(pDevCtx, pClCtx, &ioctl, sizeof(ioctl.in));
}
#endif

DfIoStatus
ioctlUnconfigure(
	Admxrc3DeviceContext* pDevCtx,
	Admxrc3ClientContext* pClCtx,
  void* pBuffer,
  unsigned int inSize)
{
#if defined(ADMXRC3_CHECK_FLAGS)
  const uint32_t legalFlagsMask = ADMXRC3_CONFIGURE_PARTIAL | ADMXRC3_CONFIGURE_SHARE;
#endif
  CoreTicket ticket;
  CoreTicket* pTicket = &ticket;
  IOCTLS_ADMXRC3_UNCONFIGURE* pIoctl = (IOCTLS_ADMXRC3_UNCONFIGURE*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  DfSpinLockFlags f;
  CoreAcquireStatus acqStatus;
  unsigned int targetIndex;
  uint32_t flags;
  boolean_t bInterrupt, bValue, bAcquired = FALSE;
  DfTime start, expires;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }
  if (!pClCtx->bPrivileged) {
    return DfIoStatusError(ADMXRC3_ACCESS_DENIED);
  }

  targetIndex = pIoctl->in.targetIndex;
  flags = pIoctl->in.flags;
  
  dfDebugPrint(6, ("ioctlUnconfigure: targetIndex=%lu flags=0x%lx\n",
    (unsigned long)targetIndex, (unsigned long)flags));

  if (targetIndex >= pDevCtx->info.numTargetFpga) {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }
#if defined(ADMXRC3_CHECK_FLAGS)
  if (flags & ~legalFlagsMask) {
    return DfIoStatusError(ADMXRC3_INVALID_FLAG);
  }
#endif

  f = dfSpinLockGet(&pDevCtx->fpga[targetIndex].owner.lock);
  if (pDevCtx->fpga[targetIndex].owner.pClCtx != NULL && pClCtx != pDevCtx->fpga[targetIndex].owner.pClCtx) {
    /* Target FPGA is currently not owned by this client; don't proceed. */
    dfSpinLockPut(&pDevCtx->fpga[targetIndex].owner.lock, f);
    return DfIoStatusError(ADMXRC3_NOT_OWNER);
  }
  dfSpinLockPut(&pDevCtx->fpga[targetIndex].owner.lock, f);

  if (!(flags & ADMXRC3_CONFIGURE_PARTIAL)) {
    pTicket->count = 1;
    pTicket->resources[0].code = CORE_RESOURCE_TARGETFPGA;
    pTicket->resources[0].index = (uint8_t)targetIndex;
    acqStatus = pDevCtx->coreInterface.pAcquireInitialize(pDevCtx->pCoreContext, pTicket);
    if (CoreAcquireSuccess != acqStatus) {
      status = mapAcquireStatus(acqStatus);
      goto done;
    }
    acqStatus = pDevCtx->coreInterface.pAcquireSync(pDevCtx->pCoreContext, pTicket, 0);
    if (CoreAcquireSuccess != acqStatus) {
      status = mapAcquireStatus(acqStatus);
      goto done;
    }
    bAcquired = TRUE;

    /* The first phase of full configuration must complete within 1s */
    start = dfTimeGet();
    expires = start + dfMicrosecondsToTime(1000000);

    /* Assert PROG */
    bValue = TRUE;
    pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpSetProg, &bValue);

    /* Wait for DONE deasserted */
    while (1) {
      pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpGetDone, &bValue);
      if (!bValue) {
        break;
      }

      bInterrupt = dfDelayThreadFor(dfMicrosecondsToTime(10000));
      if (bInterrupt) {
        /* Deassert PROG */
        bValue = FALSE;
        pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpSetProg, &bValue);

        /* Abort early */
        status = DfIoStatusError(ADMXRC3_CANCELLED);
        goto done;
      }
      if (dfTimeGet() >= expires) {
        /* Timeout waiting for DONE deasserted */
        bValue = FALSE;
        pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpSetProg, &bValue);
        status = DfIoStatusError(ADMXRC3_HARDWARE_ERROR);
        goto done;
      }
    }

    /* Wait for INIT# asserted */
    while (1) {
      pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpGetInit, &bValue);
      if (bValue) {
        break;
      }

      bInterrupt = dfDelayThreadFor(dfMicrosecondsToTime(10000));
      if (bInterrupt) {
        /* Deassert PROG */
        bValue = FALSE;
        pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpSetProg, &bValue);

        /* Abort early */
        status = DfIoStatusError(ADMXRC3_CANCELLED);
        goto done;
      }
      if (dfTimeGet() >= expires) {
        /* Timeout waiting for INIT# asserted */
        bValue = FALSE;
        pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpSetProg, &bValue);
        status = DfIoStatusError(ADMXRC3_HARDWARE_ERROR);
        goto done;
      }
    }

    /* Deassert PROG */
    bValue = FALSE;
    pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpSetProg, &bValue);

    /* Wait for INIT# deasserted */
    while (1) {
      pDevCtx->coreInterface.pFpgaControl(pDevCtx->pCoreContext, pTicket, targetIndex, CoreFpgaControlOpGetInit, &bValue);
      if (!bValue) {
        break;
      }

      bInterrupt = dfDelayThreadFor(dfMicrosecondsToTime(10000));
      if (bInterrupt) {
        /* Abort early */
        status = DfIoStatusError(ADMXRC3_CANCELLED);
        goto done;
      }
      if (dfTimeGet() >= expires) {
        /* Timeout waiting for INIT# deasserted */
        status = DfIoStatusError(ADMXRC3_HARDWARE_ERROR);
        goto done;
      }
    }
  }

done:
  if (bAcquired) {
    pDevCtx->coreInterface.pRelease(pDevCtx->pCoreContext, pTicket);
    bAcquired = FALSE;
  }

  if (!(flags & ADMXRC3_CONFIGURE_SHARE)) {
    /* Regardless of any error, mark this target FPGA as not owned by any client. */
    f = dfSpinLockGet(&pDevCtx->fpga[targetIndex].owner.lock);
    pDevCtx->fpga[targetIndex].owner.pClCtx = NULL;
    dfSpinLockPut(&pDevCtx->fpga[targetIndex].owner.lock, f);
  }

  return status;
}

